"""Classes for handling messages from the hardware device."""

from __future__ import annotations

import asyncio
import logging
import re
from typing import TYPE_CHECKING, Any

from . import const
from .const import LOCAL_CPDID
from .error import KaleidescapeError, MessageError, MessageParseError

if TYPE_CHECKING:
    from .connection import Connection

MESSAGE_TYPE_REQUEST = "request"
MESSAGE_TYPE_RESPONSE = "response"
MESSAGE_TYPE_EVENT = "event"

CPDID_FORMAT = re.compile(r"^\d\d$")
DEVICEID_FORMAT = re.compile(r"^(?P<id>\d\d|#[0-9A-F]+|\?\?)(?:\.(?P<zone>\d\d))?/")
SEQ_FORMAT = re.compile(r"^(?:\d|!)/")
STATUS_FORMAT = re.compile(r"^\d\d\d:")
NAME_FORMAT = re.compile(r"^(?:([^:]+):/?|/)")

registry = {}

_LOGGER = logging.getLogger(__name__)


def register(cls):
    """Decorator registering a message response class for factory use."""
    registry[cls.name] = cls
    return cls


class MessageParser:
    """Class for parsing messages from the hardware device."""

    def __init__(self, message: str, *args):
        """Parses string message into its fields."""
        self.message = message

        self.device_id: str = ""
        self.zone: int = 0
        self.seq: int = 0
        self.status: int = 0
        self.name: str = ""
        self.fields: list[str] = []
        self.checksum: int = 0

        is_request: bool = args[0] if len(args) > 0 else False

        pos = self._parse_device_id(0)
        pos = self._parse_seq(pos)
        if not is_request:
            pos = self._parse_status(pos)
        pos = self._parse_name(pos)
        pos = self._parse_fields(pos)
        if not is_request:
            self._parse_checksum(pos)

    def _parse_device_id(self, pos: int) -> int:
        match = re.search(DEVICEID_FORMAT, self.message[pos:])
        if not match:
            raise MessageParseError(const.ERROR_INVALID_DEVICE, self.message)
        self.device_id = match.group("id")
        self.zone = int(match.group("zone")) if match.group("zone") else 0
        return pos + match.end()

    def _parse_seq(self, pos: int) -> int:
        if not re.match(SEQ_FORMAT, self.message[pos:]):
            raise MessageParseError(const.ERROR_INVALID_SEQ_NUMBER, self.message)
        self.seq = int(self.message[pos]) if self.message[pos] != "!" else -1
        return pos + 2

    def _parse_status(self, pos: int) -> int:
        if not re.match(STATUS_FORMAT, self.message[pos:]):
            raise MessageParseError(const.ERROR_UNDETERMINED_ERROR, self.message)
        self.status = int(self.message[pos : pos + 3])
        return pos + 4

    def _parse_name(self, pos: int) -> int:
        match = re.search(NAME_FORMAT, self.message[pos:])
        if not match:
            raise MessageParseError(const.ERROR_INVALID_REQUEST, self.message)
        self.name = self.message[pos : pos + match.end(1)]
        return pos + match.end(1) + 1

    def _parse_fields(self, pos: int) -> int:
        field: str = ""
        escaped: bool = False

        while pos < len(self.message):
            char = self.message[pos]
            if escaped:
                if char == "d":
                    field += chr(int(self.message[pos + 1 : pos + 4]))
                    pos = pos + 3
                elif char == "r":
                    field += "\r"
                elif char == "n":
                    field += "\n"
                elif char == "t":
                    field += "\t"
                elif char == "/":
                    field += "/"
                elif char == "\\":
                    field += "\\"
                elif char == ":":
                    field += ":"
                elif char == "\n":
                    # Protocol bug: newline chars are not coming in encoded
                    field += "\n"
                elif char == "\r":
                    # Same bug as above
                    field += "\r"
                escaped = False
            elif char == "\\":
                escaped = True
            elif char == "/":
                break
            elif char == ":":
                self.fields.append(field)
                field = ""
            else:
                field += char

            pos = pos + 1

        if field:
            # Last field not terminated with colon
            raise MessageParseError(const.ERROR_INVALID_PARAMETER, self.message)

        return pos + 1

    def _parse_checksum(self, pos: int):
        if not re.match(r"^\d+$", self.message[pos:]):
            raise MessageParseError(const.ERROR_CHECKSUM_ERROR, self.message)
        self.checksum = int(self.message[pos:])

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"message='{self.message}', "
            f"device_id='{self.device_id}', "
            f"zone={self.zone}, "
            f"seq={self.seq}, "
            f"status={self.status}, "
            f"name='{self.name}', "
            f"fields={self.fields}, "
            f"checksum={self.checksum}"
            ")"
        )


class Message:
    """Abstract class for both request messages (command) and response messages
    (command and event)."""

    name: str = ""
    multiline = False

    def __init__(
        self,
        device_id: str = LOCAL_CPDID,
        zone: int = 0,
        seq: int = 0,
        status: int = 0,
        fields: list[str] | None = None,
    ):
        """Initializes message."""
        self._device_id = device_id
        self._zone = zone
        self._seq = seq
        self._status = status
        self._fields: list[str] = fields if fields is not None else []
        self._message: str = ""
        self._type: str = ""

    @property
    def message(self) -> str:
        """Returns body of the message."""
        return self._message

    @property
    def device_id(self) -> str:
        """The device_id of the message."""
        return self._device_id

    @property
    def zone(self) -> int:
        """Returns zone of the message."""
        return self._zone

    @property
    def seq(self) -> int:
        """Returns sequence of the message."""
        return self._seq

    @seq.setter
    def seq(self, value: int) -> None:
        """Sets sequence number of the message."""
        self._seq = value

    @property
    def status(self) -> int:
        """Returns status of the message."""
        return self._status

    @property
    def type(self) -> str:
        """Returns type of message (request/response)."""
        return self._type

    @property
    def count(self) -> int:
        """Returns how many messages are in message group.

        This is used for responses such as CONTENT_DETAILS which return multiple
        messages in a group.
        """
        raise RuntimeError("This message type does not support the 'count' property")

    def __str__(self) -> str:
        return str(self.message)


class Request(Message):
    """Class representing a command request sent to hardware device."""

    log_invalid_request: bool = True

    def __init__(self, zone: int = 0, fields: list[str] | None = None):
        """Initializes request."""
        super().__init__(
            LOCAL_CPDID, zone, fields=fields if fields is not None else fields
        )

        self.seq = -1
        self._type = MESSAGE_TYPE_REQUEST
        self._responses: list[Response] = []
        self._event = asyncio.Event()

    @property
    def fields(self) -> list[str]:
        """Returns fields in the message."""
        return self._fields

    async def send(self, connection: Connection) -> list[Response]:
        """Sends request to hardware device, returning one or more responses."""
        self._event.clear()
        self._responses.clear()

        response = await connection.send(self)

        if not response.multiline:
            connection.clear(self)

        if response.is_error:
            lvl = logging.ERROR
            if (
                not self.log_invalid_request
                and response.status == const.ERROR_INVALID_REQUEST
            ):
                lvl = logging.DEBUG
            _LOGGER.log(lvl, "Request %s failed with '%s'", repr(self), response.error)
            raise MessageError(response.status, str(self))

        _LOGGER.debug("Request %s received %s", repr(self), repr(response))

        if response.multiline:
            if not response.count:
                _LOGGER.error(
                    "Command %s response had no count '%s'", repr(self), repr(response)
                )
                raise KaleidescapeError("Response count expected")

            async def collector():
                while (len(self._responses) - 1) < response.count:
                    await asyncio.sleep(0)

            try:
                await asyncio.wait_for(collector(), connection.timeout)
            except asyncio.TimeoutError as error:
                err = f"Command {repr(self)} timed out waiting for responses"
                _LOGGER.warning(err)
                raise KaleidescapeError(err) from error

            connection.clear(self)
            return self._responses

        return [response]

    async def wait(self) -> Response:
        """Wait until the event is set."""
        await self._event.wait()
        return self._responses[0]

    def set(self, response: Response) -> None:
        """Complete request."""
        self._responses.append(response)
        self._event.set()

    def __str__(self) -> str:
        if self._message == "":
            seq = "!" if self.seq < 0 else str(self.seq)
            fields = ":".join(self._fields) + ":" if self._fields else ""
            self._message = f"{self._device_id}/{seq}/{self.name}:{fields}"
        return self._message

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"device_id={self.device_id}, "
            f"seq={self.seq}, "
            f"fields={self.fields}"
            ")"
        )


class Response(Message):
    """Class representing a command response or event from hardware device."""

    def __init__(self, parsed: MessageParser):
        """Initializes response."""
        super().__init__(
            parsed.device_id, parsed.zone, parsed.seq, parsed.status, parsed.fields
        )
        self._message = parsed.message
        self._type = MESSAGE_TYPE_EVENT if parsed.seq < 0 else MESSAGE_TYPE_RESPONSE

    @classmethod
    def factory(cls, message: str) -> Response:
        """Returns a new response object for message."""
        parsed = MessageParser(message)
        if parsed.name in registry:
            return registry[parsed.name](parsed)
        return cls(parsed)

    @property
    def is_event(self) -> bool:
        """Returns if response is an event."""
        return self._type == MESSAGE_TYPE_EVENT

    @property
    def is_error(self) -> bool:
        """Returns if response is an error."""
        return bool(self._status)

    @property
    def error(self) -> str:
        """Returns text associated with this message's status."""
        return const.RESPONSE_ERROR[self._status]

    @property
    def fields(self) -> Any:
        """Returns fields in the message."""
        return self._fields

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"device_id={self.device_id}, "
            f"seq={self.seq}, "
            f"status={self.status}, "
            f"fields={self.fields}"
            ")"
        )


@register
class Ack(Response):
    """Class for empty response messages."""

    name = ""


class GetSystemPairingInfo(Request):
    """Class for GET_SYSTEM_PAIRING_INFO messages."""

    log_invalid_request = False
    name = f"GET_{const.SYSTEM_PAIRING_INFO}"


@register
class SystemPairingInfo(Response):
    """Class for SYSTEM_PAIRING_INFO messages."""

    name = const.SYSTEM_PAIRING_INFO

    @property
    def is_paired(self) -> bool:
        """Returns if system is paired."""
        return self._fields[0] != ""

    @property
    def field_paired_system_id(self) -> str:
        """Returns system id of paired peer."""
        return self._fields[1].lower()

    @property
    def field_paired_friendly_name(self) -> str:
        """Returns friendly name of peer."""
        return self._fields[2]

    @property
    def field_paired_peers(self) -> list[tuple[str, str]]:
        """Returns list of peers."""
        if not self.is_paired:
            return []
        res: list[tuple[str, str]] = []
        for i in range(3, len(self._fields), 2):
            encore = f"{(self._fields[i].strip('#'))[-12:]:0>12}"
            premier = f"{(self._fields[i+1].strip('#'))[-12:]:0>12}"
            res.append((encore.upper(), premier.upper()))
        return res


class GetAvailableDevices(Request):
    """Class for GET_AVAILABLE_DEVICES messages."""

    name = f"GET_{const.AVAILABLE_DEVICES}"


@register
class AvailableDevices(Response):
    """Class for AVAILABLE_DEVICES messages."""

    name = const.AVAILABLE_DEVICES

    @property
    def field(self) -> list[str]:
        """Returns list of available devices."""
        return self._fields


class GetAvailableDevicesBySerialNumber(Request):
    """Class for GET_AVAILABLE_DEVICES_BY_SERIAL_NUMBER messages."""

    name = f"GET_{const.AVAILABLE_DEVICES_BY_SERIAL_NUMBER}"


@register
class AvailableDevicesBySerialNumber(Response):
    """Class for AVAILABLE_DEVICES_BY_SERIAL_NUMBER messages."""

    name = const.AVAILABLE_DEVICES_BY_SERIAL_NUMBER

    @property
    def field(self) -> list[str]:
        """Returns list of serial numbers."""
        return [f"{i[-12:]:0>12}" for i in self._fields]


class GetSystemVersion(Request):
    """Class for GET_SYSTEM_VERSION messages."""

    name = f"GET_{const.SYSTEM_VERSION}"


@register
class SystemVersion(Response):
    """Class for SYSTEM_VERSION messages."""

    name = const.SYSTEM_VERSION

    @property
    def field_protocol(self) -> int:
        """Returns protocol."""
        return int(self._fields[0])

    @property
    def field_kos(self) -> str:
        """Returns kos."""
        return self._fields[1]


class GetDeviceInfo(Request):
    """Class for GET_DEVICE_INFO messages."""

    name = f"GET_{const.DEVICE_INFO}"


@register
class DeviceInfo(Response):
    """Class for DEVICE_INFO messages."""

    name = const.DEVICE_INFO

    @property
    def field_serial_number(self) -> str:
        """Returns serial number."""
        return f"{self._fields[1][-12:]:0>12}"

    @property
    def field_cpdid(self) -> str:
        """Returns cpdid."""
        return self._fields[2] if int(self._fields[2]) else ""

    @property
    def field_ip(self) -> str:
        """Returns ip."""
        return re.sub(r"\b0+(\d)", r"\1", self._fields[3])


class GetZoneCapabilities(Request):
    """Class for GET_ZONE_CAPABILITIES messages."""

    name = f"GET_{const.ZONE_CAPABILITIES}"


@register
class ZoneCapabilities(Response):
    """Class for ZONE_CAPABILITIES messages."""

    name = const.ZONE_CAPABILITIES

    @property
    def field_osd(self) -> bool:
        """Returns whether zone has osd."""
        return self._fields[0] == "Y"

    @property
    def field_movies(self) -> bool:
        """Returns whether has movie zones."""
        return self._fields[1] == "Y"

    @property
    def field_music(self) -> bool:
        """Returns whether has music zones."""
        return self._fields[2] == "Y"

    @property
    def field_store(self) -> bool:
        """Returns whether has store."""
        return self._fields[3] == "Y"


class GetNumZones(Request):
    """Class for GET_NUM_ZONES messages."""

    name = f"GET_{const.NUM_ZONES}"


@register
class NumZones(Response):
    """Class for NUM_ZONES messages."""

    name = const.NUM_ZONES

    @property
    def field_movie_zones(self) -> int:
        """Returns movie zones."""
        return int(self._fields[0])

    @property
    def field_music_zones(self) -> int:
        """Returns music zones."""
        return int(self._fields[1])


class GetDeviceTypeName(Request):
    """Class for GET_DEVICE_TYPE_NAME messages."""

    name = f"GET_{const.DEVICE_TYPE_NAME}"


@register
class DeviceTypeName(Response):
    """Class for DEVICE_TYPE_NAME messages."""

    name = const.DEVICE_TYPE_NAME

    @property
    def field(self) -> str:
        """Returns device type name."""
        return self._fields[0]


class GetDevicePowerState(Request):
    """Class for GET_DEVICE_POWER_STATE messages."""

    name = f"GET_{const.DEVICE_POWER_STATE}"


@register
class DevicePowerState(Response):
    """Class for DEVICE_POWER_STATE messages."""

    name = const.DEVICE_POWER_STATE
    index = {
        "power": {
            0: const.DEVICE_POWER_STATE_STANDBY,
            1: const.DEVICE_POWER_STATE_ON,
        },
        "zone": {
            0: const.DEVICE_ZONE_STATE_DISABLED,
            1: const.DEVICE_ZONE_STATE_AVAILABLE,
        },
    }

    @property
    def field_power(self) -> str:
        """Returns power state."""
        return self.index["power"][int(self._fields[0])]

    @property
    def field_zone(self) -> list[str]:
        """Returns zone state."""
        return [self.index["zone"][int(v)] for v in self._fields[1:]]


class GetSystemReadinessState(Request):
    """Class for GET_SYSTEM_READINESS_STATE messages."""

    name = f"GET_{const.SYSTEM_READINESS_STATE}"


@register
class SystemReadinessState(Response):
    """Class for SYSTEM_READINESS_STATE messages."""

    name = const.SYSTEM_READINESS_STATE
    index = {
        0: const.SYSTEM_READINESS_STATE_READY,
        1: const.SYSTEM_READINESS_STATE_BECOMING_READY,
        2: const.SYSTEM_READINESS_STATE_IDLE,
    }

    @property
    def field(self) -> str:
        """Returns readiness state."""
        return self.index[int(self._fields[0])]


class GetPlayStatus(Request):
    """Class for GET_PLAY_STATUS messages."""

    name = f"GET_{const.PLAY_STATUS}"


@register
class PlayStatus(Response):
    """Class for PLAY_STATUS messages."""

    name = const.PLAY_STATUS
    index = {
        0: const.PLAY_STATUS_NONE,
        1: const.PLAY_STATUS_PAUSED,
        2: const.PLAY_STATUS_PLAYING,
        4: const.PLAY_STATUS_FORWARD,
        6: const.PLAY_STATUS_REVERSE,
    }

    @property
    def field_play_status(self) -> str:
        """Return play status."""
        return self.index[int(self._fields[0])]

    @property
    def field_play_speed(self) -> int:
        """Returns play speed."""
        return int(self._fields[1])

    @property
    def field_title_number(self) -> int:
        """Returns title number."""
        return int(self._fields[2])

    @property
    def field_title_length(self) -> int:
        """Returns title length."""
        return int(self._fields[3])

    @property
    def field_title_location(self) -> int:
        """Returns title location."""
        return int(self._fields[4])

    @property
    def field_chapter_number(self) -> int:
        """Returns chapter number."""
        return int(self._fields[5])

    @property
    def field_chapter_length(self) -> int:
        """Returns chapter length."""
        return int(self._fields[6])

    @property
    def field_chapter_location(self) -> int:
        """Returns chapter location."""
        return int(self._fields[7])


class GetFriendlySystemName(Request):
    """Class for GET_FRIENDLY_SYSTEM_NAME messages."""

    name = f"GET_{const.FRIENDLY_SYSTEM_NAME}"


@register
class FriendlySystemName(Response):
    """Class for FRIENDLY_SYSTEM_NAME messages."""

    name = const.FRIENDLY_SYSTEM_NAME

    @property
    def field(self) -> str:
        """Returns friendly system name."""
        return self._fields[0]


class GetFriendlyName(Request):
    """Class for GET_FRIENDLY_NAME messages."""

    name = f"GET_{const.FRIENDLY_NAME}"


@register
class FriendlyName(Response):
    """Class for FRIENDLY_NAME messages."""

    name = const.FRIENDLY_NAME

    @property
    def field(self) -> str:
        """Returns friendly name."""
        return self._fields[0]


class GetUiState(Request):
    """Class for GET_UI_STATE messages."""

    name = f"GET_{const.UI_STATE}"


@register
class UiState(Response):
    """Class for UI_STATE messages."""

    name = const.UI_STATE
    index = {
        const.UI_STATE_SCREEN: {
            0: const.UI_STATE_SCREEN_UNKNOWN,
            1: const.UI_STATE_SCREEN_MOVIE_LIST,
            2: const.UI_STATE_SCREEN_MOVIE_COLLECTIONS,
            3: const.UI_STATE_SCREEN_MOVIE_COVERS,
            4: const.UI_STATE_SCREEN_PARENTAL_CONTROL,
            7: const.UI_STATE_SCREEN_PLAYING_MOVIE,
            8: const.UI_STATE_SCREEN_SYSTEM_STATUS,
            9: const.UI_STATE_SCREEN_MUSIC_LIST,
            10: const.UI_STATE_SCREEN_MUSIC_COVERS,
            11: const.UI_STATE_SCREEN_MUSIC_COLLECTIONS,
            12: const.UI_STATE_SCREEN_MUSIC_NOW_PLAYING,
            14: const.UI_STATE_SCREEN_VAULT_SUMMARY,
            15: const.UI_STATE_SCREEN_SYSTEM_SETTINGS,
            16: const.UI_STATE_SCREEN_MOVIE_STORE,
            17: const.UI_STATE_SCREEN_PAIRED_UNIT_LOBBY,
        },
        const.UI_STATE_POPUP: {
            0: const.UI_STATE_POPUP_NONE,
            1: const.UI_STATE_POPUP_DETAILS,
            2: const.UI_STATE_POPUP_MOVIE_STATUS,
            3: const.UI_STATE_POPUP_MOVIE_NOT_STATUS,
        },
        const.UI_STATE_DIALOG: {
            0: const.UI_STATE_DIALOG_NONE,
            1: const.UI_STATE_DIALOG_MENU,
            2: const.UI_STATE_DIALOG_PASSCODE,
            3: const.UI_STATE_DIALOG_QUESTION,
            4: const.UI_STATE_DIALOG_INFORMATION,
            5: const.UI_STATE_DIALOG_WARNING,
            6: const.UI_STATE_DIALOG_ERROR,
            7: const.UI_STATE_DIALOG_PREPLAY,
            8: const.UI_STATE_DIALOG_WARRANTY,
            9: const.UI_STATE_DIALOG_KEYBOARD,
            10: const.UI_STATE_DIALOG_IP_CONFIG,
        },
        const.UI_STATE_SAVER: {
            0: const.UI_STATE_SAVER_INACTIVE,
            1: const.UI_STATE_SAVER_ACTIVE,
        },
    }

    @property
    def field_screen(self) -> str:
        """Returns screen."""
        return self.index[const.UI_STATE_SCREEN][int(self._fields[0])]

    @property
    def field_popup(self) -> str:
        """Returns popup."""
        return self.index[const.UI_STATE_POPUP][int(self._fields[1])]

    @property
    def field_dialog(self) -> str:
        """Returns dialog."""
        return self.index[const.UI_STATE_DIALOG][int(self._fields[2])]

    @property
    def field_screensaver(self) -> str:
        """Returns screensaver."""
        return self.index[const.UI_STATE_SAVER][int(self._fields[3])]


class GetPlayingTitleName(Request):
    """Class for GET_PLAYING_TITLE_NAME messages."""

    name = f"GET_{const.PLAYING_TITLE_NAME}"


@register
class PlayingTitleName(Response):
    """Class for TITLE_NAME messages."""

    name = const.TITLE_NAME

    @property
    def field(self) -> str:
        """Returns title name."""
        return self._fields[0]


class GetHighlightedSelection(Request):
    """Class for GET_HIGHLIGHTED_SELECTION messages."""

    name = f"GET_{const.HIGHLIGHTED_SELECTION}"


@register
class HighlightedSelection(Response):
    """Class for HIGHLIGHTED_SELECTION messages."""

    name = const.HIGHLIGHTED_SELECTION

    @property
    def field(self) -> str:
        """Returns highlighted selection."""
        return self._fields[0]


class GetContentDetails(Request):
    """Class for GET_CONTENT_DETAILS messages."""

    name = f"GET_{const.CONTENT_DETAILS}"


@register
class ContentDetailsOverview(Response):
    """Class for CONTENT_DETAILS_OVERVIEW messages."""

    name = const.CONTENT_DETAILS_OVERVIEW
    multiline = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.details: dict[str, str] = {}

    @property
    def count(self) -> int:
        """Returns number of content detail rows."""
        return int(self._fields[0])

    @property
    def field_handle(self) -> str:
        """Returns handle."""
        return self._fields[1]

    @property
    def field_table(self) -> str:
        """Returns table."""
        return self._fields[2]

    @property
    def field_title(self) -> str:
        """Returns title."""
        return self.details["Title"] if "Title" in self.details else ""

    @property
    def field_cover_url(self) -> str:
        """Returns cover url."""
        if "Cover_URL" in self.details:
            return self.details["Cover_URL"]
        return ""

    @property
    def field_hires_cover_url(self) -> str:
        """Returns hires cover url."""
        if "HiRes_cover_URL" in self.details:
            return self.details["HiRes_cover_URL"]
        return ""

    @property
    def field_rating(self) -> str:
        """Returns rating."""
        if "Rating" in self.details:
            return self.details["Rating"]
        return ""

    @property
    def field_rating_reason(self) -> str:
        """Returns rating reason."""
        if "Rating_reason" in self.details:
            return self.details["Rating_reason"]
        return ""

    @property
    def field_year(self) -> str:
        """Returns year."""
        if "Year" in self.details:
            return self.details["Year"]
        return ""

    @property
    def field_running_time(self) -> str:
        """Returns running time."""
        if "Running_time" in self.details:
            return self.details["Running_time"]
        return "0"

    @property
    def field_actors(self) -> list[str]:
        """Returns actors."""
        if "Actors" in self.details:
            return re.split("[\n\r]+", self.details["Actors"])
        return []

    @property
    def field_director(self) -> str:
        """Returns director."""
        if "Director" in self.details:
            return self.details["Director"]
        return ""

    @property
    def field_directors(self) -> list[str]:
        """Returns directors."""
        if "Directors" in self.details:
            return re.split("[\n\r]+", self.details["Directors"])
        return []

    @property
    def field_genre(self) -> str:
        """Returns genre."""
        if "Genre" in self.details:
            return self.details["Genre"]
        return ""

    @property
    def field_genres(self) -> list[str]:
        """Returns genres."""
        if "Genres" in self.details:
            return re.split("[\n\r]+", self.details["Genres"])
        return []

    @property
    def field_synopsis(self) -> str:
        """Returns synopsis."""
        if "Synopsis" in self.details:
            return self.details["Synopsis"]
        return ""

    @property
    def field_color_description(self) -> str:
        """Returns color description."""
        if "Color_description" in self.details:
            return self.details["Color_description"]
        return ""

    @property
    def field_country(self) -> str:
        """Returns country."""
        if "Country" in self.details:
            return self.details["Country"]
        return ""

    @property
    def field_aspect_ratio(self) -> str:
        """Returns aspect ratio."""
        if "Aspect_ratio" in self.details:
            return self.details["Aspect_ratio"]
        return ""


@register
class ContentDetails(Response):
    """Class for CONTENT_DETAILS messages."""

    name = const.CONTENT_DETAILS

    @property
    def field(self) -> dict[str, str]:
        """Returns content details."""
        return {self.fields[1]: self.fields[2]}


class GetMovieLocation(Request):
    """Class for GET_MOVIE_LOCATION messages."""

    name = f"GET_{const.MOVIE_LOCATION}"


@register
class MovieLocation(Response):
    """Class for MOVIE_LOCATION messages."""

    name = const.MOVIE_LOCATION
    index = {
        0: const.MOVIE_LOCATION_NONE,
        3: const.MOVIE_LOCATION_CONTENT,
        4: const.MOVIE_LOCATION_INTERMISSION,
        5: const.MOVIE_LOCATION_CREDITS,
        6: const.MOVIE_LOCATION_DISC_MENU,
    }

    @property
    def field(self) -> str:
        """Returns movie location."""
        return self.index[int(self._fields[0])]


class GetMovieMediaType(Request):
    """Class for GET_MOVIE_MEDIA_TYPE messages."""

    name = f"GET_{const.MOVIE_MEDIA_TYPE}"


@register
class MovieMediaType(Response):
    """Class for MOVIE_MEDIA_TYPE messages."""

    name = const.MOVIE_MEDIA_TYPE
    index = {
        0: const.MOVIE_MEDIA_TYPE_NONE,
        1: const.MOVIE_MEDIA_TYPE_DVD,
        2: const.MOVIE_MEDIA_TYPE_STREAM,
        3: const.MOVIE_MEDIA_TYPE_BLURAY,
    }

    @property
    def field(self) -> str:
        """Returns media type."""
        return self.index[int(self._fields[0])]


class GetVideoColor(Request):
    """Class for GET_VIDEO_COLOR messages."""

    name = f"GET_{const.VIDEO_COLOR}"


@register
class VideoColor(Response):
    """Class for VIDEO_COLOR messages."""

    name = const.VIDEO_COLOR
    index = {
        const.VIDEO_COLOR_EOTF: {
            0: const.VIDEO_COLOR_EOTF_UNKNOWN,
            1: const.VIDEO_COLOR_EOTF_SDR,
            2: const.VIDEO_COLOR_EOTF_HDR,
            3: const.VIDEO_COLOR_EOTF_SMTPEST2084,
        },
        const.VIDEO_COLOR_SPACE: {
            0: const.VIDEO_COLOR_SPACE_DEFAULT,
            1: const.VIDEO_COLOR_SPACE_RGB,
            2: const.VIDEO_COLOR_SPACE_BT601,
            3: const.VIDEO_COLOR_SPACE_BT709,
            4: const.VIDEO_COLOR_SPACE_BT2020,
        },
        const.VIDEO_COLOR_DEPTH: {
            0: const.VIDEO_COLOR_DEPTH_UNKNOWN,
            24: const.VIDEO_COLOR_DEPTH_24BIT,
            30: const.VIDEO_COLOR_DEPTH_30BIT,
            36: const.VIDEO_COLOR_DEPTH_36BIT,
        },
        const.VIDEO_COLOR_SAMPLING: {
            0: const.VIDEO_COLOR_SAMPLING_NONE,
            1: const.VIDEO_COLOR_SAMPLING_RGB,
            2: const.VIDEO_COLOR_SAMPLING_YCBCR422,
            3: const.VIDEO_COLOR_SAMPLING_YCBCR444,
            4: const.VIDEO_COLOR_SAMPLING_YCBCR420,
        },
    }

    @property
    def field_eotf(self) -> str:
        """Returns eotf."""
        return self.index[const.VIDEO_COLOR_EOTF][int(self._fields[0])]

    @property
    def field_space(self) -> str:
        """Returns space."""
        return self.index[const.VIDEO_COLOR_SPACE][int(self._fields[1])]

    @property
    def field_depth(self) -> str:
        """Returns depth."""
        return self.index[const.VIDEO_COLOR_DEPTH][int(self._fields[2])]

    @property
    def field_sampling(self) -> str:
        """Returns sampling."""
        return self.index[const.VIDEO_COLOR_SAMPLING][int(self._fields[3])]


class GetVideoMode(Request):
    """Class for GET_VIDEO_MODE messages."""

    name = f"GET_{const.VIDEO_MODE}"


@register
class VideoMode(Response):
    """Class for VIDEO_MODE messages."""

    name = const.VIDEO_MODE
    index = {
        0: const.VIDEO_MODE_NONE,
        1: const.VIDEO_MODE_480I60_4X3,
        2: const.VIDEO_MODE_480I60_16X9,
        3: const.VIDEO_MODE_480P60_4X3,
        4: const.VIDEO_MODE_480P60_16X9,
        5: const.VIDEO_MODE_576I50_4X3,
        6: const.VIDEO_MODE_576I50_16X9,
        7: const.VIDEO_MODE_576P50_4X3,
        8: const.VIDEO_MODE_576P50_16X9,
        9: const.VIDEO_MODE_720P60_NTSC_HD,
        10: const.VIDEO_MODE_720P50_PAL_HD,
        11: const.VIDEO_MODE_1080I60_16X9,
        12: const.VIDEO_MODE_1080I50_16X9,
        13: const.VIDEO_MODE_1080P60_16X9,
        14: const.VIDEO_MODE_1080P50_16X9,
        17: const.VIDEO_MODE_1080P24_16X9,
        19: const.VIDEO_MODE_480I60_64X27,
        20: const.VIDEO_MODE_576I50_64X27,
        21: const.VIDEO_MODE_1080I60_64X27,
        22: const.VIDEO_MODE_1080I50_64X27,
        23: const.VIDEO_MODE_1080P60_64X27,
        24: const.VIDEO_MODE_1080P50_64X27,
        25: const.VIDEO_MODE_1080P23976_64X27,
        26: const.VIDEO_MODE_1080P24_64X27,
        27: const.VIDEO_MODE_3840X2160P23976_16X9,
        28: const.VIDEO_MODE_3840X2160P23976_64X27,
        29: const.VIDEO_MODE_3840X2160P30_16X9,
        30: const.VIDEO_MODE_3840X2160P30_64X27,
        31: const.VIDEO_MODE_3840X2160P60_16X9,
        32: const.VIDEO_MODE_3840X2160P60_64X27,
        33: const.VIDEO_MODE_3840X2160P25_16X9,
        34: const.VIDEO_MODE_3840X2160P25_64X27,
        35: const.VIDEO_MODE_3840X2160P50_16X9,
        36: const.VIDEO_MODE_3840X2160P50_64X27,
        37: const.VIDEO_MODE_3840X2160P24_16X9,
        38: const.VIDEO_MODE_3840X2160P24_64X27,
    }

    @property
    def field(self) -> str:
        """Returns video mode."""
        return self.index[int(self._fields[2])]


class GetScreenMask(Request):
    """Class for GET_SCREEN_MASK messages."""

    name = f"GET_{const.SCREEN_MASK}"


@register
class ScreenMask(Response):
    """Class for SCREEN_MASK messages."""

    name = const.SCREEN_MASK
    index = {
        0: const.SCREEN_MASK_ASPECT_RATIO_NONE,
        1: const.SCREEN_MASK_ASPECT_RATIO_133,
        2: const.SCREEN_MASK_ASPECT_RATIO_166,
        3: const.SCREEN_MASK_ASPECT_RATIO_178,
        4: const.SCREEN_MASK_ASPECT_RATIO_185,
        5: const.SCREEN_MASK_ASPECT_RATIO_235,
    }

    @property
    def field_image_ratio(self) -> str:
        """Returns image ratio."""
        return self.index[int(self._fields[0])]

    @property
    def field_top_trim_rel(self) -> int:
        """Returns top trim rel."""
        return int(self._fields[1])

    @property
    def field_bottom_trim_rel(self) -> int:
        """Returns bottom trim rel."""
        return int(self._fields[2])

    @property
    def field_conservative_ratio(self) -> str:
        """Returns conservative ratio."""
        return self.index[int(self._fields[3])]

    @property
    def field_top_mask_abs(self) -> int:
        """Returns top mask abs."""
        return int(self._fields[4])

    @property
    def field_bottom_mask_abs(self) -> int:
        """Returns bottom mask abs."""
        return int(self._fields[5])


class GetScreenMask2(Request):
    """Class for GET_SCREEN_MASK2 messages."""

    name = f"GET_{const.SCREEN_MASK2}"


@register
class ScreenMask2(Response):
    """Class for SCREEN_MASK2 messages."""

    name = const.SCREEN_MASK2

    @property
    def field_top_mask_abs(self) -> int:
        """Returns top mask abs."""
        return int(self._fields[0])

    @property
    def field_bottom_mask_abs(self) -> int:
        """Returns bottom mask abs."""
        return int(self._fields[1])

    @property
    def field_top_calibrated(self) -> int:
        """Returns top calibrated."""
        return int(self._fields[2])

    @property
    def field_bottom_calibrated(self) -> int:
        """Returns bottom calibrated."""
        return int(self._fields[3])


class GetCinemascapeMode(Request):
    """Class for GET_CINEMASCAPE_MODE messages."""

    name = f"GET_{const.CINEMASCAPE_MODE}"


@register
class CinemascapeMode(Response):
    """Class for CINEMASCAPE_MODE messages."""

    name = const.CINEMASCAPE_MODE
    index = {
        0: const.CINEMASCAPE_MODE_NONE,
        1: const.CINEMASCAPE_MODE_ANAMORPHIC,
        2: const.CINEMASCAPE_MODE_LETTERBOX,
        3: const.CINEMASCAPE_MODE_NATIVE,
    }

    @property
    def field(self) -> str:
        """Returns cinemascape mode."""
        return self.index[int(self._fields[0])]


class GetCinemascapeMask(Request):
    """Class for GET_CINEMASCAPE_MASK messages."""

    name = f"GET_{const.CINEMASCAPE_MASK}"


@register
class CinemascapeMask(Response):
    """Class for CINEMASCAPE_MASK messages."""

    name = const.CINEMASCAPE_MASK

    @property
    def field(self) -> int:
        """Returns cinemascape mask."""
        return int(self._fields[0])


class EnableEvents(Request):
    """Class for ENABLE_EVENTS messages."""

    name = const.ENABLE_EVENTS


class LeaveStandby(Request):
    """Class for LEAVE_STANDBY messages."""

    name = const.LEAVE_STANDBY


class EnterStandby(Request):
    """Class for ENTER_STANDBY messages."""

    name = const.ENTER_STANDBY


class Play(Request):
    """Class for PLAY messages."""

    name = const.PLAY


class Pause(Request):
    """Class for PAUSE messages."""

    name = const.PAUSE


class Stop(Request):
    """Class for STOP messages."""

    name = const.STOP


class Next(Request):
    """Class for NEXT messages."""

    name = const.NEXT


class Previous(Request):
    """Class for PREVIOUS messages."""

    name = const.PREVIOUS


class Replay(Request):
    """Class for REPLAY messages."""

    name = const.REPLAY


class ScanForward(Request):
    """Class for SCAN_FORWARD messages."""

    name = const.SCAN_FORWARD


class ScanReverse(Request):
    """Class for SCAN_REVERSE messages."""

    name = const.SCAN_REVERSE


class Select(Request):
    """Class for SELECT messages."""

    name = const.SELECT


class Up(Request):
    """Class for UP messages."""

    name = const.UP


class Down(Request):
    """Class for DOWN messages."""

    name = const.DOWN


class Left(Request):
    """Class for LEFT messages."""

    name = const.LEFT


class Right(Request):
    """Class for RIGHT messages."""

    name = const.RIGHT


class Cancel(Request):
    """Class for CANCEL messages."""

    name = const.CANCEL


class GoMovieCovers(Request):
    """Class for GO_MOVIE_LIST messages."""

    name = const.GO_MOVIE_COVERS


class MenuToggle(Request):
    """Class for KALEIDESCAPE_MENU_TOGGLE messages."""

    name = const.KALEIDESCAPE_MENU_TOGGLE



"""Unit tests."""

from __future__ import annotations

import asyncio
import logging

from kaleidescape.dispatcher import Dispatcher

_LOGGER = logging.getLogger(__name__)


def create_signal(dispatcher: Dispatcher, target_event: str) -> asyncio.Event:
    """Returns an asyncio event that is triggered when event is emitted."""
    trigger = asyncio.Event()

    async def handler(event):
        if event == target_event:
            trigger.set()

    dispatcher.connect(handler)

    return trigger




"""Tests for connection module."""
import asyncio
from unittest.mock import MagicMock, patch

import dns.exception
import pytest
from dns.rdtypes.IN.A import A

from kaleidescape.const import STATE_CONNECTED, STATE_DISCONNECTED, STATE_RECONNECTING
from kaleidescape.connection import Connection
from kaleidescape.dispatcher import Dispatcher

from . import create_signal
from .emulator import Emulator

# pylint: disable=unused-argument


def test_init():
    """Test init sets properties."""
    connection = Connection(Dispatcher())
    assert isinstance(connection.dispatcher, Dispatcher)


@pytest.mark.asyncio
async def test_connect_fails():
    """Test connect fails when host not available."""
    connection = Connection(Dispatcher())
    with pytest.raises(ConnectionError) as err:
        await connection.connect("127.0.0.1", port=10001, timeout=1)
    assert isinstance(err.value, ConnectionRefusedError)

    # Also fails for initial connection even with reconnect.
    with pytest.raises(ConnectionError) as err:
        await connection.connect("127.0.0.1", port=10001, timeout=1, reconnect=True)
    assert isinstance(err.value, ConnectionRefusedError)


@pytest.mark.asyncio
async def test_connect_timeout():
    """Test connect fails when host not available."""
    connection = Connection(Dispatcher())
    with pytest.raises(ConnectionError):
        await connection.connect("0.0.0.1", port=10001, timeout=1)

    # Also fails for initial connection even with reconnect.
    with pytest.raises(ConnectionError):
        await connection.connect("0.0.0.1", port=10001, timeout=1, reconnect=True)


@pytest.mark.asyncio
async def test_connect_succeeds(emulator: Emulator):
    """Test connect updates state and emits signal."""
    dispatcher = Dispatcher()
    connection = Connection(dispatcher)
    assert connection.state == STATE_DISCONNECTED
    signal = create_signal(dispatcher, STATE_CONNECTED)
    await connection.connect("127.0.0.1", port=10001, timeout=1)
    await signal.wait()
    assert connection.ip == "127.0.0.1"
    assert connection.port == 10001
    assert connection.timeout == 1
    assert connection.state == STATE_CONNECTED
    await connection.disconnect()


@pytest.mark.asyncio
async def test_manual_disconnect(emulator: Emulator, connection: Connection):
    """Test disconnect updates state and emits signal."""
    signal = create_signal(connection.dispatcher, STATE_DISCONNECTED)
    await connection.disconnect()
    await signal.wait()
    assert connection.state == STATE_DISCONNECTED


@pytest.mark.asyncio
async def test_event_disconnect(emulator: Emulator, connection: Connection):
    """Test connection error during event results in disconnected."""
    signal = create_signal(connection.dispatcher, STATE_DISCONNECTED)
    await emulator.stop()
    await signal.wait()
    assert connection.state == STATE_DISCONNECTED


@pytest.mark.asyncio
async def test_reconnect_during_event(emulator: Emulator):
    """Test reconnect while waiting for events/responses."""
    dispatcher = Dispatcher()
    connection = Connection(dispatcher)

    connect_signal = create_signal(dispatcher, STATE_CONNECTED)
    disconnect_signal = create_signal(dispatcher, STATE_DISCONNECTED)

    # Assert connection
    await connection.connect(
        "127.0.0.1", port=10001, timeout=1, reconnect=True, reconnect_delay=1
    )
    await connect_signal.wait()
    assert connection.state == STATE_CONNECTED
    connect_signal.clear()

    # Assert transitions to reconnecting and emits disconnect signal
    await emulator.stop()
    await disconnect_signal.wait()
    assert connection.state == STATE_RECONNECTING

    # Assert reconnects once server is back up and emits connected signal
    await emulator.start()
    await connect_signal.wait()
    assert connection.state == STATE_CONNECTED

    await connection.disconnect()


@pytest.mark.asyncio
async def test_reconnect_cancelled(emulator):
    """Test reconnect is canceled by calling disconnect."""
    dispatcher = Dispatcher()
    connection = Connection(dispatcher)

    connect_signal = create_signal(dispatcher, STATE_CONNECTED)
    disconnect_signal = create_signal(dispatcher, STATE_DISCONNECTED)

    # Assert open and fires connected
    await connection.connect(
        "127.0.0.1", port=10001, timeout=1, reconnect=True, reconnect_delay=0.5
    )
    await connect_signal.wait()
    assert connection.state == STATE_CONNECTED

    # Assert transitions to reconnecting and emits disconnect signal
    await emulator.stop()
    await disconnect_signal.wait()
    assert connection.state == STATE_RECONNECTING

    await connection.disconnect()
    assert connection.state == STATE_DISCONNECTED


@pytest.mark.asyncio
async def test_resolve_succeeds(emulator: Emulator):
    """Test resolve when host name resolution succeeds."""
    with patch("dns.asyncresolver.Resolver.resolve") as mock:
        mock.return_value = [MagicMock(spec=A)]
        mock.return_value[0].to_text.side_effect = [
            "127.0.0.1",  # 1st call: mDSN succeeds
            dns.exception.DNSException,  # 2nd call: mDSN fails
            "127.0.0.1"  # 2nd call: DSN succeeds
        ]
        # First call simulates mDNS lookup
        assert await Connection.resolve("my-kaleidescape") == "127.0.0.1"
        # Second call simulates DNS lookup
        assert await Connection.resolve("some-kaleidescape") == "127.0.0.1"


@pytest.mark.asyncio
async def test_resolve_fails(emulator: Emulator):
    """Test resolve when host name resolution fails."""
    with patch("dns.asyncresolver.Resolver.resolve") as mock:
        mock.return_value = [MagicMock(spec=A)]
        mock.return_value[0].to_text.side_effect = dns.exception.DNSException
        with pytest.raises(ConnectionError) as err:
            assert await Connection.resolve("my-kaleidescape") == "127.0.0.1"
        assert isinstance(err.value, ConnectionError)
        assert str(err.value) == "Failed to resolve host my-kaleidescape"



"""Class handling network connection to hardware device."""

from __future__ import annotations

import asyncio
import logging
import re
from collections.abc import Callable
from typing import TYPE_CHECKING

import dns.asyncresolver
import dns.exception

from . import const
from .error import KaleidescapeError, MessageParseError, format_error
from .message import Response

if TYPE_CHECKING:
    import dns.resolver
    from dns.rdtypes.IN.A import A

    from .dispatcher import Dispatcher
    from .message import Request

_LOGGER = logging.getLogger(__name__)

SEPARATOR = "\n"
SEPARATOR_BYTES = SEPARATOR.encode("latin-1")


class Connection:
    """Class handling network connection to hardware device."""

    def __init__(self, dispatcher: Dispatcher, on_event: Callable = None) -> None:
        """Initializes connection."""
        self._dispatcher = dispatcher
        self._on_event = on_event

        self._ip: str | None = None
        self._port: int | None = None
        self._timeout: float | None = None
        self._state: str = const.STATE_DISCONNECTED
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None
        self._response_handler_task: asyncio.Task | None = None
        self._reconnect_delay: float | None = None
        self._reconnect_task: asyncio.Task | None = None
        self._reconnect_enabled: bool = False
        self._pending_requests: dict[int, Request] = {}

    @property
    def dispatcher(self) -> Dispatcher:
        """Return dispatcher instance."""
        return self._dispatcher

    @property
    def ip(self) -> str | None:
        """Return ip of the server connected to."""
        return self._ip

    @property
    def port(self) -> int | None:
        """Return port of the hardware device to connect to."""
        return self._port

    @property
    def timeout(self) -> float | None:
        """Return connection timeout in seconds."""
        return self._timeout

    @property
    def state(self) -> str:
        """Return state of the connection to the hardware device."""
        return self._state

    async def connect(
        self,
        ip: str,
        port: int,
        timeout: float,
        reconnect: bool = False,
        reconnect_delay: float = const.DEFAULT_RECONNECT_DELAY,
    ) -> None:
        """Connect to the hardware device."""
        if self._state == const.STATE_CONNECTED:
            return

        self._ip = ip
        self._port = port
        self._timeout = timeout

        # Disable auto_reconnect until a good connect
        self._reconnect_enabled = False

        await self._connect()

        self._reconnect_enabled = reconnect
        self._reconnect_delay = reconnect_delay

        _LOGGER.info("Connected to %s", self._ip)

    async def _connect(self) -> None:
        """Connect to server."""
        try:
            connection = asyncio.open_connection(self._ip, self._port)
            self._reader, self._writer = await asyncio.wait_for(
                connection, self._timeout
            )
        except ConnectionError:
            # Don't allow subclasses of ConnectionError to be cast as OSErrors below
            raise
        except (OSError, asyncio.TimeoutError) as err:
            # Generalize connection errors
            raise ConnectionError(format_error(err)) from err

        self._response_handler_task = asyncio.create_task(self._response_handler())

        self._state = const.STATE_CONNECTED
        self._dispatcher.send(const.STATE_CONNECTED)

    async def _response_handler(self) -> None:
        """Main loop receiving responses and events from hardware device."""
        assert self._reader

        while True:
            try:
                result = await self._reader.readuntil()

                response = Response.factory(result.decode("latin-1").strip())
                _LOGGER.debug("Response received '%s'", response.message)

                if response.is_event:
                    # Events are unsolicited notifications about a change in state.
                    if self._on_event:
                        asyncio.create_task(self._on_event(response))
                elif response.device_id == const.LOCAL_CPDID:
                    if response.seq not in self._pending_requests:
                        _LOGGER.error("Response seq not registered '%s'", response)
                    else:
                        request = self._pending_requests[response.seq]
                        request.set(response)
            except (asyncio.IncompleteReadError, OSError) as err:
                asyncio.create_task(self._handle_connection_error(err))
                return
            except MessageParseError as err:
                _LOGGER.exception(err)
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.exception(
                    "Unhandled exception %s('%s')", type(err).__name__, err
                )

    async def _handle_connection_error(self, err: Exception):
        """Handle connection failures and schedule reconnect."""
        if self._reconnect_task:
            return

        await self._disconnect()

        if self._reconnect_enabled:
            self._state = const.STATE_RECONNECTING
            self._reconnect_task = asyncio.create_task(self._reconnect())
        else:
            self._state = const.STATE_DISCONNECTED

        _LOGGER.error(
            "Disconnected from %s %s('%s')", self._ip, type(err).__name__, err
        )
        self._dispatcher.send(const.STATE_DISCONNECTED)

    async def _reconnect(self):
        """Reconnect to server."""
        try:
            while self._state != const.STATE_CONNECTED:
                try:
                    await self._connect()
                except ConnectionError as err:
                    _LOGGER.warning("Failed reconnect to %s with '%s'", self._ip, err)
                    await self._disconnect()
                    await asyncio.sleep(self._reconnect_delay)
                else:
                    self._reconnect_task = None
                    _LOGGER.info("Reconnected to %s", self._ip)
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.exception("Unhandled exception %s('%s')", type(err).__name__, err)
            raise

    async def disconnect(self):
        """Disconnect from server."""
        if self._state == const.STATE_DISCONNECTED:
            return

        # Cancel pending reconnect task
        if self._reconnect_task:
            self._reconnect_task.cancel()
            try:
                await self._reconnect_task
            except:  # pylint: disable=bare-except
                # Ensure completes
                pass
            self._reconnect_task = None

        await self._disconnect()
        self._state = const.STATE_DISCONNECTED

        _LOGGER.info("Disconnected from %s", self._ip)
        self._dispatcher.send(const.STATE_DISCONNECTED)

    async def _disconnect(self):
        """Disconnect from server."""
        if self._response_handler_task:
            self._response_handler_task.cancel()
            try:
                await self._response_handler_task
            except:  # pylint: disable=bare-except
                # Ensure completes
                pass
            self._response_handler_task = None

        if self._writer:
            self._writer.close()
            self._writer = None

        self._reader = None
        self._pending_requests.clear()

    async def send(self, request: Request) -> Response:
        """Send request to device"""
        if self._state != const.STATE_CONNECTED:
            err = "Not connected to device"
            _LOGGER.error(err)
            raise KaleidescapeError(err)

        wait = 0.01
        retries = self.timeout * (1/wait)

        while request.seq < 0:
            try:
                # Devices can only handle 10 concurrent requests. Find next available
                # sequence number not in use.
                request.seq = next(
                    (i for i in range(0, 10) if i not in self._pending_requests)
                )
            except StopIteration:
                await asyncio.sleep(wait)
                retries = retries - 1
                if retries == 0:
                    raise ConnectionError
                continue

        self._pending_requests[request.seq] = request

        try:
            assert self._writer
            writer = self._writer
            writer.write(str(request).encode("latin-1") + SEPARATOR_BYTES)
            await writer.drain()
            _LOGGER.debug("Request sent '%s'", request)
            response = await asyncio.wait_for(request.wait(), self._timeout)
        except (OSError, ConnectionError, asyncio.TimeoutError) as err:
            msg = f"Request '{request}' failed with '{format_error(err)}'"
            _LOGGER.warning(msg)
            raise KaleidescapeError(msg) from err

        return response

    def clear(self, request: Request):
        """Clear request from the pending requests list, indicating response has been
        received."""
        if request.seq not in self._pending_requests:
            _LOGGER.error("Request seq not registered '%s'", request)
        else:
            self._pending_requests.pop(request.seq)

    @staticmethod
    async def resolve(host: str, timeout: int = 5) -> str:
        """Resolve hostname to ip address."""

        async def _resolve(use_mdns: bool) -> str:
            resolver = dns.asyncresolver.Resolver()
            resolver.lifetime = timeout
            if use_mdns:
                resolver.nameservers = ["224.0.0.251"]
                resolver.port = 5353
            answer: list[A] = await resolver.resolve(host, rdtype="A")
            if len(answer) == 0:
                raise RuntimeError("Answer expected")
            return answer[0].to_text()

        ip_address = host

        if re.search("^[0-9.]+$", host) is None:
            try:
                # Attempt resolving via mDNS
                ip_address = await _resolve(True)
            except dns.exception.DNSException:
                try:
                    # Attempt resolving via DNS
                    ip_address = await _resolve(False)
                except dns.exception.DNSException as err:
                    raise ConnectionError(f"Failed to resolve host {host}") from err
        else:
            # Normalize IP by removing leading zeros
            ip_address = re.sub(r"\b0+(\d)", r"\1", ip_address)

        if ip_address != host:
            _LOGGER.debug("Resolved %s to %s", host, ip_address)

        return ip_address



"""Custom errors."""
from __future__ import annotations

import asyncio

from . import const

DEFAULT_MESSAGES = {
    asyncio.TimeoutError: "Command timed out",
    ConnectionError: "Connection error",
    BrokenPipeError: "Broken pipe",
    ConnectionAbortedError: "Connection aborted",
    ConnectionRefusedError: "Connection refused",
    ConnectionResetError: "Connection reset",
    OSError: "OS I/O error",
}


def format_error(err: Exception | asyncio.TimeoutError) -> str:
    """Formats error message based on a base error."""
    msg: str | None = str(err)
    if msg == "":
        msg = DEFAULT_MESSAGES.get(type(err))
    return msg if msg else ""


class KaleidescapeError(Exception):
    """Kaleidescape errors."""


class SystemNotFoundError(Exception):
    """Error finding system."""


class MessageError(KaleidescapeError, RuntimeError):
    """Errors from the Kaleidescape Control Protocol."""

    def __init__(self, code: int, message: str = None):
        self.code = code
        self.error = const.RESPONSE_ERROR[code]
        super().__init__(self.error + (f" for command '{message}'" if message else ""))


class MessageParseError(KaleidescapeError, ValueError):
    """Errors parsing Kaleidescape Control Protocol messages."""

    def __init__(self, code: int, message: str):
        self.code = code
        self.error = const.RESPONSE_ERROR[code]
        super().__init__(self.error + (f" for command '{message}'" if message else ""))



"""pytest fixtures"""

import asyncio

import pytest

from kaleidescape.connection import Connection
from kaleidescape.dispatcher import Dispatcher
from tests.emulator import Emulator


@pytest.fixture(name="emulator")
def fixture_emulator(event_loop):
    """Fixture for creating a device emulator."""
    emulator = Emulator("127.0.0.1", port=10001)
    event_loop.run_until_complete(emulator.start())
    yield emulator
    event_loop.run_until_complete(emulator.stop())


@pytest.fixture(name="connection")
def fixture_connection(event_loop):
    """Fixture for creating a connection to device."""
    connection = Connection(Dispatcher())
    event_loop.run_until_complete(
        connection.connect("127.0.0.1", port=10001, timeout=1)
    )
    yield connection
    # Ensure work emulator loop can complete tasks
    event_loop.run_until_complete(asyncio.sleep(0.01))
    event_loop.run_until_complete(connection.disconnect())




"""Constants."""

LOCAL_CPDID = "01"

# Defaults
DEFAULT_PROTOCOL_PORT = 10000
DEFAULT_PROTOCOL_TIMEOUT = 10.0
DEFAULT_RECONNECT_DELAY = 10.0

# Connection
STATE_CONNECTED = "connected"
STATE_DISCONNECTED = "disconnected"
STATE_RECONNECTING = "reconnecting"

# Control Protocol Statuses
SUCCESS = 0
ERROR_MESSAGE_TO_LONG = 1
ERROR_MESSAGE_INVALID_CHARACTERS = 2
ERROR_CHECKSUM_ERROR = 3
ERROR_INVALID_DEVICE = 4
ERROR_DEVICE_UNAVAILABLE = 5
ERROR_INVALID_ZONE_SYNTAX = 6
ERROR_INVALID_ZONE = 7
ERROR_INVALID_REQUEST = 10
ERROR_INVALID_NUMBER_PARAMETERS = 11
ERROR_INVALID_PARAMETER = 12
ERROR_DEVICE_ID_CONFLICT = 13
ERROR_INVALID_SEQ_NUMBER = 14
ERROR_INVALID_PASSCODE = 16
ERROR_INVALID_CONTENT_HANDLE = 17
ERROR_NETWORK_ERROR = 18
ERROR_INVALID_SERIAL_SYNTAX = 19
ERROR_DEVICE_IN_STANDBY = 20
ERROR_REJECTED_WHILE_PAIRED = 32

ERROR_UNDETERMINED_ERROR = 999
RESPONSE_ERROR = {
    SUCCESS: "Success",
    ERROR_MESSAGE_TO_LONG: "Message too long",
    ERROR_MESSAGE_INVALID_CHARACTERS: "Message contains invalid characters",
    ERROR_CHECKSUM_ERROR: "Checksum error",
    ERROR_INVALID_DEVICE: "Invalid device",
    ERROR_DEVICE_UNAVAILABLE: "Device unavailable",
    ERROR_INVALID_ZONE_SYNTAX: "Invalid zone syntax",
    ERROR_INVALID_ZONE: "Invalid zone",
    ERROR_INVALID_REQUEST: "Invalid request",
    ERROR_INVALID_NUMBER_PARAMETERS: "Invalid number of parameters",
    ERROR_INVALID_PARAMETER: "Invalid parameter",
    ERROR_DEVICE_ID_CONFLICT: "Device identifier conflict",
    ERROR_INVALID_SEQ_NUMBER: "Invalid sequence number",
    ERROR_INVALID_PASSCODE: "Invalid passcode",
    ERROR_INVALID_CONTENT_HANDLE: "Invalid content handle",
    ERROR_NETWORK_ERROR: "Network error",
    ERROR_INVALID_SERIAL_SYNTAX: "Invalid serial syntax",
    ERROR_DEVICE_IN_STANDBY: "Device in standby",
    ERROR_REJECTED_WHILE_PAIRED: "Command is rejected while in paired state",
    ERROR_UNDETERMINED_ERROR: "Undetermined error",
}

#
# Control Protocol Messages
#

SYSTEM_PAIRING_INFO = "SYSTEM_PAIRING_INFO"
SYSTEM_VERSION = "SYSTEM_VERSION"
DEVICE_INFO = "DEVICE_INFO"
ZONE_CAPABILITIES = "ZONE_CAPABILITIES"
NUM_ZONES = "NUM_ZONES"
DEVICE_TYPE_NAME = "DEVICE_TYPE_NAME"
FRIENDLY_SYSTEM_NAME = "FRIENDLY_SYSTEM_NAME"
FRIENDLY_NAME = "FRIENDLY_NAME"
AVAILABLE_DEVICES = "AVAILABLE_DEVICES"
AVAILABLE_DEVICES_BY_SERIAL_NUMBER = "AVAILABLE_DEVICES_BY_SERIAL_NUMBER"
PLAYING_TITLE_NAME = "PLAYING_TITLE_NAME"
TITLE_NAME = "TITLE_NAME"
HIGHLIGHTED_SELECTION = "HIGHLIGHTED_SELECTION"
CONTENT_DETAILS = "CONTENT_DETAILS"
CONTENT_DETAILS_OVERVIEW = "CONTENT_DETAILS_OVERVIEW"
ENABLE_EVENTS = "ENABLE_EVENTS"
ENTER_STANDBY = "ENTER_STANDBY"
LEAVE_STANDBY = "LEAVE_STANDBY"
PLAY = "PLAY"
PAUSE = "PAUSE"
STOP = "STOP"
NEXT = "NEXT"
PREVIOUS = "PREVIOUS"
REPLAY = "REPLAY"
SCAN_FORWARD = "SCAN_FORWARD"
SCAN_REVERSE = "SCAN_REVERSE"
SELECT = "SELECT"
UP = "UP"
DOWN = "DOWN"
LEFT = "LEFT"
RIGHT = "RIGHT"
CANCEL = "CANCEL"
GO_MOVIE_COVERS = "GO_MOVIE_COVERS"
KALEIDESCAPE_MENU_TOGGLE = "KALEIDESCAPE_MENU_TOGGLE"

DEVICE_POWER_STATE = "DEVICE_POWER_STATE"
DEVICE_POWER_STATE_STANDBY = "standby"
DEVICE_POWER_STATE_ON = "on"
DEVICE_ZONE_STATE_DISABLED = "disabled"
DEVICE_ZONE_STATE_AVAILABLE = "available"

SYSTEM_READINESS_STATE = "SYSTEM_READINESS_STATE"
SYSTEM_READINESS_STATE_READY = "ready"
SYSTEM_READINESS_STATE_BECOMING_READY = "becoming_ready"
SYSTEM_READINESS_STATE_IDLE = "idle"

PLAY_STATUS = "PLAY_STATUS"
PLAY_STATUS_NONE = "none"
PLAY_STATUS_PAUSED = "paused"
PLAY_STATUS_PLAYING = "playing"
PLAY_STATUS_FORWARD = "forward"
PLAY_STATUS_REVERSE = "reverse"

MOVIE_LOCATION = "MOVIE_LOCATION"
MOVIE_LOCATION_NONE = "none"
MOVIE_LOCATION_CONTENT = "content"
MOVIE_LOCATION_INTERMISSION = "intermission"
MOVIE_LOCATION_CREDITS = "credits"
MOVIE_LOCATION_DISC_MENU = "disc_menu"

MOVIE_MEDIA_TYPE = "MOVIE_MEDIA_TYPE"
MOVIE_MEDIA_TYPE_NONE = "none"
MOVIE_MEDIA_TYPE_DVD = "dvd"
MOVIE_MEDIA_TYPE_STREAM = "stream"
MOVIE_MEDIA_TYPE_BLURAY = "bluray"

UI_STATE = "UI_STATE"
UI_STATE_SCREEN = "screen"
UI_STATE_SCREEN_UNKNOWN = "unknown"
UI_STATE_SCREEN_MOVIE_LIST = "movie_list"
UI_STATE_SCREEN_MOVIE_COLLECTIONS = "movie_collections"
UI_STATE_SCREEN_MOVIE_COVERS = "movie_covers"
UI_STATE_SCREEN_PARENTAL_CONTROL = "parental_control"
UI_STATE_SCREEN_PLAYING_MOVIE = "playing_movie"
UI_STATE_SCREEN_SYSTEM_STATUS = "system_status"
UI_STATE_SCREEN_MUSIC_LIST = "music_list"
UI_STATE_SCREEN_MUSIC_COVERS = "music_covers"
UI_STATE_SCREEN_MUSIC_COLLECTIONS = "music_collections"
UI_STATE_SCREEN_MUSIC_NOW_PLAYING = "music_now_playing"
UI_STATE_SCREEN_VAULT_SUMMARY = "vault_summary"
UI_STATE_SCREEN_SYSTEM_SETTINGS = "system_settings"
UI_STATE_SCREEN_MOVIE_STORE = "movie_store"
UI_STATE_SCREEN_PAIRED_UNIT_LOBBY = "paired_unit_lobby"
UI_STATE_POPUP = "popup"
UI_STATE_POPUP_NONE = "none"
UI_STATE_POPUP_DETAILS = "details"
UI_STATE_POPUP_MOVIE_STATUS = "movie_status"
UI_STATE_POPUP_MOVIE_NOT_STATUS = "movie_not_status"
UI_STATE_DIALOG = "dialog"
UI_STATE_DIALOG_NONE = "none"
UI_STATE_DIALOG_MENU = "menu"
UI_STATE_DIALOG_PASSCODE = "passcode"
UI_STATE_DIALOG_QUESTION = "question"
UI_STATE_DIALOG_INFORMATION = "information"
UI_STATE_DIALOG_WARNING = "warning"
UI_STATE_DIALOG_ERROR = "error"
UI_STATE_DIALOG_PREPLAY = "preplay"
UI_STATE_DIALOG_WARRANTY = "warranty"
UI_STATE_DIALOG_KEYBOARD = "keyboard"
UI_STATE_DIALOG_IP_CONFIG = "ip_config"
UI_STATE_SAVER = "saver"
UI_STATE_SAVER_INACTIVE = "inactive"
UI_STATE_SAVER_ACTIVE = "active"

VIDEO_MODE = "VIDEO_MODE"
VIDEO_MODE_NONE = "none"
VIDEO_MODE_480I60_4X3 = "480i60_4:3"
VIDEO_MODE_480I60_16X9 = "480i60_16:9"
VIDEO_MODE_480P60_4X3 = "480p60_4:3"
VIDEO_MODE_480P60_16X9 = "480p60_16:9"
VIDEO_MODE_576I50_4X3 = "576i50_4:3"
VIDEO_MODE_576I50_16X9 = "576i50_16:9"
VIDEO_MODE_576P50_4X3 = "576p50_4:3"
VIDEO_MODE_576P50_16X9 = "576p50_16:9"
VIDEO_MODE_720P60_NTSC_HD = "720p60_ntsc_hd"
VIDEO_MODE_720P50_PAL_HD = "720p50_pal_hd"
VIDEO_MODE_1080I60_16X9 = "1080i60_16:9"
VIDEO_MODE_1080I50_16X9 = "1080i50_16:9"
VIDEO_MODE_1080P60_16X9 = "1080p60_16:9"
VIDEO_MODE_1080P50_16X9 = "1080p50_16:9"
VIDEO_MODE_1080P24_16X9 = "1080p24_16:9"
VIDEO_MODE_480I60_64X27 = "480i60_64:27"
VIDEO_MODE_576I50_64X27 = "576i50_64:27"
VIDEO_MODE_1080I60_64X27 = "1080i60_64:27"
VIDEO_MODE_1080I50_64X27 = "1080i50_64:27"
VIDEO_MODE_1080P60_64X27 = "1080p60_64:27"
VIDEO_MODE_1080P50_64X27 = "1080p50_64:27"
VIDEO_MODE_1080P23976_64X27 = "1080p23976_64:27"
VIDEO_MODE_1080P24_64X27 = "1080p24_64:27"
VIDEO_MODE_3840X2160P23976_16X9 = "3840x2160p23976_16:9"
VIDEO_MODE_3840X2160P23976_64X27 = "3840x2160p23976_64:27"
VIDEO_MODE_3840X2160P30_16X9 = "3840x2160p30_16:9"
VIDEO_MODE_3840X2160P30_64X27 = "3840x2160p30_64:27"
VIDEO_MODE_3840X2160P60_16X9 = "3840x2160p60_16:9"
VIDEO_MODE_3840X2160P60_64X27 = "3840x2160p60_64:27"
VIDEO_MODE_3840X2160P25_16X9 = "3840x2160p25_16:9"
VIDEO_MODE_3840X2160P25_64X27 = "3840x2160p25_64:27"
VIDEO_MODE_3840X2160P50_16X9 = "3840x2160p50_16:9"
VIDEO_MODE_3840X2160P50_64X27 = "3840x2160p50_64:27"
VIDEO_MODE_3840X2160P24_16X9 = "3840x2160p24_16:9"
VIDEO_MODE_3840X2160P24_64X27 = "3840x2160p24_64:27"

VIDEO_COLOR = "VIDEO_COLOR"
VIDEO_COLOR_EOTF = "eotf"
VIDEO_COLOR_EOTF_UNKNOWN = "unknown"
VIDEO_COLOR_EOTF_SDR = "sdr"
VIDEO_COLOR_EOTF_HDR = "hdr"
VIDEO_COLOR_EOTF_SMTPEST2084 = "smtpest2084"
VIDEO_COLOR_SPACE = "space"
VIDEO_COLOR_SPACE_DEFAULT = "default"
VIDEO_COLOR_SPACE_RGB = "rgb"
VIDEO_COLOR_SPACE_BT601 = "bt601"
VIDEO_COLOR_SPACE_BT709 = "bt709"
VIDEO_COLOR_SPACE_BT2020 = "bt2020"
VIDEO_COLOR_DEPTH = "depth"
VIDEO_COLOR_DEPTH_UNKNOWN = "unknown"
VIDEO_COLOR_DEPTH_24BIT = "24bit"
VIDEO_COLOR_DEPTH_30BIT = "30bit"
VIDEO_COLOR_DEPTH_36BIT = "36bit"
VIDEO_COLOR_SAMPLING = "sampling"
VIDEO_COLOR_SAMPLING_NONE = "none"
VIDEO_COLOR_SAMPLING_RGB = "rgb"
VIDEO_COLOR_SAMPLING_YCBCR422 = "ycbcr422"
VIDEO_COLOR_SAMPLING_YCBCR444 = "ycbcr444"
VIDEO_COLOR_SAMPLING_YCBCR420 = "ycbcr420"

SCREEN_MASK = "SCREEN_MASK"
SCREEN_MASK_ASPECT_RATIO_NONE = "none"
SCREEN_MASK_ASPECT_RATIO_133 = "1.33"
SCREEN_MASK_ASPECT_RATIO_166 = "1.66"
SCREEN_MASK_ASPECT_RATIO_178 = "1.78"
SCREEN_MASK_ASPECT_RATIO_185 = "1.85"
SCREEN_MASK_ASPECT_RATIO_235 = "2.35"

SCREEN_MASK2 = "SCREEN_MASK2"

CINEMASCAPE_MODE = "CINEMASCAPE_MODE"
CINEMASCAPE_MODE_NONE = "none"
CINEMASCAPE_MODE_ANAMORPHIC = "anamorphic"
CINEMASCAPE_MODE_LETTERBOX = "letterbox"
CINEMASCAPE_MODE_NATIVE = "native"

CINEMASCAPE_MASK = "CINEMASCAPE_MASK"



"""Tests for device module."""

import asyncio

import pytest

from kaleidescape import const
from kaleidescape import message as messages
from kaleidescape.connection import Connection
from kaleidescape.const import LOCAL_CPDID, SUCCESS
from kaleidescape.device import Device
from kaleidescape.dispatcher import Dispatcher
from kaleidescape.error import KaleidescapeError, MessageError

from . import create_signal
from .emulator import Emulator

# pylint: disable=unused-argument


@pytest.mark.asyncio
async def test_init(emulator: Emulator):
    """Test initialization of device."""
    device = Device("127.0.0.1", port=10001)
    assert isinstance(device.connection, Connection)
    assert isinstance(device.dispatcher, Dispatcher)
    assert device.host == "127.0.0.1"
    assert device.port == 10001
    assert device.is_connected is False
    assert device.serial_number == ""


@pytest.mark.asyncio
async def test_connect(emulator: Emulator):
    """Test initialization of device."""
    device = Device("127.0.0.1", port=10001)
    await device.connect()
    assert device.is_connected

    await device.disconnect()
    assert not device.is_connected


@pytest.mark.asyncio
async def test_get_available_devices(emulator: Emulator):
    """Test get available devices."""
    device = Device("127.0.0.1", port=10001)
    await device.connect()
    fields = await device.get_available_devices()
    assert fields == [LOCAL_CPDID]
    await device.disconnect()


@pytest.mark.asyncio
async def test_get_available_serial_numbers(
    emulator: Emulator
):
    """Test get available devices."""
    device = Device("127.0.0.1", port=10001)
    await device.connect()
    fields = await device.get_available_serial_numbers()
    assert fields == ["00000000123A"]
    await device.disconnect()


@pytest.mark.asyncio
async def test_refresh(emulator: Emulator):
    """Test refreshing system updates state."""
    device = Device("127.0.0.1", port=10001)
    await device.refresh()
    assert device.system.serial_number == "00000000123A"
    assert device.serial_number == "00000000123A"
    assert device.system.cpdid == ""
    assert device.system.ip_address == "127.0.0.1"
    assert device.system.protocol == 16
    assert device.system.kos_version == "10.4.2-19218"
    assert device.system.type == "Strato S"
    assert device.system.friendly_name == "Theater"
    assert device.system.movie_zones == 1
    assert device.system.music_zones == 1
    assert device.power.state == const.DEVICE_POWER_STATE_STANDBY
    assert device.power.readiness == const.SYSTEM_READINESS_STATE_IDLE
    assert device.is_server_only is False
    assert device.is_movie_player is True
    await device.disconnect()


@pytest.mark.asyncio
async def test_refresh_state(emulator: Emulator):
    """Test refreshing state updates state."""
    device = Device("127.0.0.1", port=10001)
    await device.refresh()

    assert device.power.state == const.DEVICE_POWER_STATE_STANDBY
    assert device.osd.ui_screen == const.UI_STATE_SCREEN_UNKNOWN
    assert device.automation.movie_location == const.MOVIE_LOCATION_NONE

    signal = create_signal(
        device.dispatcher, messages.DevicePowerState.name
    )

    emulator.register_mock_command(
        ("01",),
        messages.GetDevicePowerState.name,
        (SUCCESS, messages.DevicePowerState.name, ["1", "1"]),
    )

    await emulator.send_event(
        [LOCAL_CPDID], SUCCESS, messages.DevicePowerState.name, ["1"]
    )

    await asyncio.wait_for(signal.wait(), 1)

    assert device.power.state == const.DEVICE_POWER_STATE_ON
    assert device.osd.ui_screen == const.UI_STATE_SCREEN_MOVIE_LIST
    assert device.automation.movie_location == const.MOVIE_LOCATION_CONTENT

    await device.disconnect()


@pytest.mark.asyncio
async def test_events(emulator: Emulator):
    """Test device events."""
    device = Device("127.0.0.1", port=10001)
    await device.connect()
    # Setup to listen for signal emit
    signal = create_signal(
        device.dispatcher, messages.SystemReadinessState.name
    )
    # Produce a signal
    await emulator.send_event(
        [LOCAL_CPDID], SUCCESS, messages.SystemReadinessState.name, ["0"]
    )
    # Assert signal is received
    await asyncio.wait_for(signal.wait(), 1)

    await device.disconnect()


@pytest.mark.asyncio
async def test_commands(emulator: Emulator):
    """Test commands."""
    device = Device("127.0.0.1", port=10001)
    await device.connect()
    await device.leave_standby()
    await device.play()
    await device.pause()
    await device.stop()
    await device.enter_standby()

    await device.disconnect()


@pytest.mark.asyncio
async def test_get_content_details(emulator: Emulator):
    """Test command involving multiline response."""
    device = Device("127.0.0.1", port=10001)
    await device.refresh()
    res = await device.get_content_details("26-0.0-S_c446c8e2")
    assert res.field_handle == "26-0.0-S_c446c8e2"
    assert res.field_table == "movies"
    assert res.field_title == "Turtle Odyssey"

    await device.disconnect()



"""Tests for command module."""
from __future__ import annotations

import asyncio
from typing import cast

import pytest

from kaleidescape import const
from kaleidescape import message as messages
from kaleidescape.connection import Connection

from kaleidescape.const import LOCAL_CPDID, STATE_CONNECTED, STATE_DISCONNECTED
from kaleidescape.dispatcher import Dispatcher
from kaleidescape.error import KaleidescapeError, MessageError

from . import create_signal
from .emulator import Emulator

# pylint: disable=unused-argument


@pytest.mark.asyncio
async def test_command_succeeds(emulator: Emulator, connection: Connection):
    """Test command succeeds."""
    req = messages.GetAvailableDevices()
    res = cast(messages.AvailableDevices, (await req.send(connection))[0])
    assert res.field == [LOCAL_CPDID]


@pytest.mark.asyncio
async def test_command_fails(emulator: Emulator, connection: Connection):
    """Test command fails."""
    emulator.register_mock_command(
        ("01",),
        messages.GetAvailableDevices.name,
        (const.ERROR_DEVICE_UNAVAILABLE, messages.AvailableDevices.name),
    )
    with pytest.raises(MessageError) as err:
        req = messages.GetAvailableDevices()
        await req.send(connection)
    assert const.RESPONSE_ERROR[const.ERROR_DEVICE_UNAVAILABLE] in str(err.value)


@pytest.mark.asyncio
async def test_commands_fail_when_disconnected(emulator: Emulator):
    """Test calling commands fail when disconnected."""
    dispatcher = Dispatcher()
    connection = Connection(dispatcher)
    await connection.connect("127.0.0.1", port=10001, timeout=1)
    await connection.disconnect()
    assert connection.state == const.STATE_DISCONNECTED
    with pytest.raises(KaleidescapeError) as err:
        req = messages.GetAvailableDevices()
        await req.send(connection)
    assert str(err.value) == "Not connected to device"


@pytest.mark.asyncio
async def test_reconnect_during_command(emulator: Emulator):
    """Test reconnect while waiting for events/responses."""
    dispatcher = Dispatcher()
    connection = Connection(dispatcher)

    connect_signal = create_signal(dispatcher, STATE_CONNECTED)
    disconnect_signal = create_signal(dispatcher, STATE_DISCONNECTED)

    # Assert connection
    await connection.connect(
        "127.0.0.1", port=10001, timeout=1, reconnect=True, reconnect_delay=1
    )
    await connect_signal.wait()
    assert connection.state == const.STATE_CONNECTED
    connect_signal.clear()

    # Break connection
    await emulator.stop()
    await emulator.start()

    # Assert command with orphaned connection times out
    with pytest.raises(KaleidescapeError) as err:
        req = messages.GetAvailableDevices()
        await req.send(connection)
    assert "Not connected to device" in str(err.value)

    # Wait for reconnect
    await disconnect_signal.wait()
    await connect_signal.wait()

    # Assert reconnected
    assert connection.state == const.STATE_CONNECTED

    await connection.disconnect()


@pytest.mark.asyncio
async def test_get_available_devices(emulator: Emulator, connection: Connection):
    """Test command."""
    req = messages.GetAvailableDevices()
    res = cast(messages.AvailableDevices, (await req.send(connection))[0])
    assert res.field == [LOCAL_CPDID]


@pytest.mark.asyncio
async def test_get_available_devices_by_serial_number(
    emulator: Emulator, connection: Connection
):
    """Test command."""
    req = messages.GetAvailableDevicesBySerialNumber()
    res = cast(messages.AvailableDevicesBySerialNumber, (await req.send(connection))[0])
    assert res.field == ["00000000123A"]


@pytest.mark.asyncio
async def test_get_device_info(emulator: Emulator, connection: Connection):
    """Test command with single device."""
    req = messages.GetDeviceInfo()
    res = cast(messages.DeviceInfo, (await req.send(connection))[0])
    assert res.field_serial_number == "00000000123A"
    assert res.field_cpdid == ""
    assert res.field_ip == "127.0.0.1"


@pytest.mark.asyncio
async def test_concurrency(emulator: Emulator, connection: Connection):
    """Test command concurrency handling."""
    requests: list[messages.GetAvailableDevices] = []
    for _ in range(0, 12):
        requests.append(messages.GetAvailableDevices())

    responses = await asyncio.gather(*[r.send(connection) for r in requests])

    assert len([r[0].seq for r in responses if r[0].seq == 0]) == 2
    assert len([r[0].seq for r in responses if r[0].seq == 1]) == 2
    assert len([r[0].seq for r in responses if r[0].seq == 2]) == 1
    assert len([r[0].seq for r in responses if r[0].seq == 9]) == 1

    await connection.disconnect()



"""Classes providing virtualized access to hardware devices."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar, cast

from . import const
from . import message as messages
from .connection import Connection
from .dispatcher import Dispatcher

if TYPE_CHECKING:
    from .dispatcher import Signal
    from .message import Request, Response

    RequestType = TypeVar("RequestType", bound=Request)


class Device:
    """Class representing hardware.

    Provide commands for changing the state of the device. Also handles mirroring
    device state by monitoring system events.
    """

    def __init__(
        self,
        host: str,
        *,
        port: int = const.DEFAULT_PROTOCOL_PORT,
        timeout: float = const.DEFAULT_PROTOCOL_TIMEOUT,
        reconnect: bool = True,
        reconnect_delay: float = const.DEFAULT_RECONNECT_DELAY,
    ) -> None:
        """Initialize device."""
        self._host = host
        self._port = port
        self._timeout = timeout
        self._reconnect_enabled = reconnect
        self._reconnect_delay = reconnect_delay

        self._dispatcher = Dispatcher()
        self._connection = Connection(self._dispatcher, self._handle_event)

        self.system = System()
        self.power = Power()
        self.osd = OSD()
        self.movie = Movie()
        self.automation = Automation()

        self._signal: Signal | None = None

    async def connect(self) -> None:
        """Connect to hardware."""
        if self.is_connected:
            return

        # Convert hostname to ip (if not already)
        self._host = await Connection.resolve(self._host)

        await self._connection.connect(
            self._host,
            port=self._port,
            timeout=self._timeout,
            reconnect=self._reconnect_enabled,
            reconnect_delay=self._reconnect_delay,
        )

        result = iter(
            await asyncio.gather(
                self._get_device_info(),
                self._get_system_version(),
                self._get_device_type_name(),
                self._get_num_zones(),
                self._get_device_power_state(),
                self._get_system_readiness_state(),
            )
        )

        self._update_device_info(next(result))
        self._update_system_version(next(result))
        self._update_device_type_name(next(result))
        self._update_num_zones(next(result))
        self._update_device_power_state(next(result))
        self._update_system_readiness_state(next(result))

        if self.is_movie_player:
            # Server only devices don't support this call
            self._update_friendly_name(await self._get_friendly_name())

    async def disconnect(self) -> None:
        """Disconnect from hardware."""
        if not self.is_connected:
            return

        await self._connection.disconnect()

    async def refresh(self) -> None:
        """Sync device state."""
        if not self.is_connected:
            await self.connect()

        if self.power.state != const.DEVICE_POWER_STATE_ON:
            return

        result = iter(
            await asyncio.gather(
                self._get_ui_state(),
                self._get_highlighted_selection(),
                self._get_play_status(),
                self._get_movie_location(),
                self._get_screen_mask(),
                self._get_screen_mask2(),
                self._get_cinemascape_mode(),
            )
        )

        self._update_ui_state(next(result))
        self._update_highlighted_selection(next(result))
        self._update_play_status(next(result))
        self._update_movie_location(next(result))
        self._update_screen_mask(next(result))
        self._update_screen_mask2(next(result))
        self._update_cinemascape_mode(next(result))

        if self.movie.play_status != const.PLAY_STATUS_NONE:
            res1 = await self.get_content_details(self.osd.highlighted)
            self._update_content_details(cast(messages.ContentDetailsOverview, res1))

        if self.automation.cinemascape_mode != const.CINEMASCAPE_MODE_NONE:
            res2 = await self._get_cinemascape_mask()
            self._update_cinemascape_mask(cast(messages.CinemascapeMask, res2))

    async def get_system_pairing_info(self) -> messages.SystemPairingInfo:
        """Return a list the serial numbers in the system."""
        res = await self._send(messages.GetSystemPairingInfo)
        return cast(messages.SystemPairingInfo, res)

    async def get_friendly_system_name(self) -> str:
        """Return friendly system name."""
        res = await self._send(messages.GetFriendlySystemName)
        return cast(messages.FriendlySystemName, res).field

    async def get_available_serial_numbers(self) -> list[str]:
        """Return a list the serial numbers in the system."""
        res = await self._send(messages.GetAvailableDevicesBySerialNumber)
        return (cast(messages.AvailableDevicesBySerialNumber, res)).field

    async def get_available_devices(self) -> list[str]:
        """Return a list of cpdid's in the system that have cpdid's assigned."""
        res = await self._send(messages.GetAvailableDevices)
        return (cast(messages.AvailableDevices, res)).field

    async def leave_standby(self) -> None:
        """Send leave standby command."""
        await self._send(messages.LeaveStandby)

    async def enter_standby(self) -> None:
        """Send enter standby command."""
        await self._send(messages.EnterStandby)

    async def play(self) -> None:
        """Send play command."""
        await self._send(messages.Play)

    async def pause(self) -> None:
        """Send pause command."""
        await self._send(messages.Pause)

    async def stop(self) -> None:
        """Send stop command."""
        await self._send(messages.Stop)

    async def next(self) -> None:
        """Send next command."""
        await self._send(messages.Next)

    async def previous(self) -> None:
        """Send previous command."""
        await self._send(messages.Previous)

    async def replay(self) -> None:
        """Send replay command."""
        await self._send(messages.Replay)

    async def scan_forward(self) -> None:
        """Send scan_forward command."""
        await self._send(messages.ScanForward)

    async def scan_reverse(self) -> None:
        """Send scan_reverse command."""
        await self._send(messages.ScanReverse)

    async def select(self) -> None:
        """Send select command."""
        await self._send(messages.Select)

    async def up(self) -> None:  # pylint: disable=invalid-name
        """Send up command."""
        await self._send(messages.Up)

    async def down(self) -> None:
        """Send down command."""
        await self._send(messages.Down)

    async def left(self) -> None:
        """Send left command."""
        await self._send(messages.Left)

    async def right(self) -> None:
        """Send right command."""
        await self._send(messages.Right)

    async def cancel(self) -> None:
        """Send cancel command."""
        await self._send(messages.Cancel)

    async def go_movie_covers(self) -> None:
        """Send list command."""
        await self._send(messages.GoMovieCovers)

    async def menu_toggle(self) -> None:
        """Send menu toggle command."""
        await self._send(messages.MenuToggle)

    async def _get_device_info(self) -> messages.DeviceInfo:
        """Return device info."""
        res = await self._send(messages.GetDeviceInfo)
        return cast(messages.DeviceInfo, res)

    def _update_device_info(self, res: messages.DeviceInfo) -> None:
        self.system.serial_number = res.field_serial_number
        self.system.cpdid = res.field_cpdid
        self.system.ip_address = res.field_ip

    async def _get_system_version(self) -> messages.SystemVersion:
        """Return system version."""
        res = await self._send(messages.GetSystemVersion)
        return cast(messages.SystemVersion, res)

    def _update_system_version(self, res: messages.SystemVersion) -> None:
        self.system.protocol = res.field_protocol
        self.system.kos_version = res.field_kos

    async def _get_num_zones(self) -> messages.NumZones:
        """Return number of zones."""
        res = await self._send(messages.GetNumZones)
        return cast(messages.NumZones, res)

    def _update_num_zones(self, res: messages.NumZones) -> None:
        self.system.movie_zones = res.field_movie_zones
        self.system.music_zones = res.field_music_zones

    async def _get_device_type_name(self) -> messages.DeviceTypeName:
        """Return device type name."""
        res = await self._send(messages.GetDeviceTypeName)
        return cast(messages.DeviceTypeName, res)

    def _update_device_type_name(self, res: messages.DeviceTypeName) -> None:
        self.system.type = res.field

    async def _get_device_power_state(self) -> messages.DevicePowerState:
        """Return power state."""
        res = await self._send(messages.GetDevicePowerState)
        return cast(messages.DevicePowerState, res)

    def _update_device_power_state(self, res: messages.DevicePowerState) -> None:
        self.power.state = res.field_power
        self.power.zone = res.field_zone

    async def _get_system_readiness_state(self) -> messages.SystemReadinessState:
        """Return readiness state."""
        res = await self._send(messages.GetSystemReadinessState)
        return cast(messages.SystemReadinessState, res)

    def _update_system_readiness_state(
        self, res: messages.SystemReadinessState
    ) -> None:
        self.power.readiness = res.field

    async def _get_friendly_name(self) -> messages.FriendlyName:
        """Return friendly name."""
        res = await self._send(messages.GetFriendlyName)
        return cast(messages.FriendlyName, res)

    def _update_friendly_name(self, res: messages.FriendlyName) -> None:
        self.system.friendly_name = res.field

    async def _get_ui_state(self) -> messages.UiState:
        """Return ui state."""
        res = await self._send(messages.GetUiState)
        return cast(messages.UiState, res)

    def _update_ui_state(self, res: messages.UiState) -> None:
        self.osd.ui_screen = res.field_screen
        self.osd.ui_popup = res.field_popup
        self.osd.ui_dialog = res.field_dialog
        self.osd.ui_screensaver = res.field_screensaver

    async def _get_playing_title_name(self) -> messages.PlayingTitleName:
        """Return playing title name."""
        res = await self._send(messages.GetPlayingTitleName)
        return cast(messages.PlayingTitleName, res)

    def _update_playing_title_name(self, res: messages.PlayingTitleName) -> None:
        self.osd.title_name = res.field

    async def _get_highlighted_selection(self) -> messages.HighlightedSelection:
        """Return highlighted selection."""
        res = await self._send(messages.GetHighlightedSelection)
        return cast(messages.HighlightedSelection, res)

    def _update_highlighted_selection(self, res: messages.HighlightedSelection) -> None:
        self.osd.highlighted = res.field

    async def _get_play_status(self) -> messages.PlayStatus:
        """Return play status."""
        res = await self._send(messages.GetPlayStatus)
        return cast(messages.PlayStatus, res)

    def _update_play_status(self, res: messages.PlayStatus) -> None:
        self.movie.play_status = res.field_play_status
        self.movie.play_speed = res.field_play_speed
        self.movie.title_number = res.field_title_number
        self.movie.title_length = res.field_title_length
        self.movie.title_location = res.field_title_location
        self.movie.chapter_number = res.field_chapter_number
        self.movie.chapter_length = res.field_chapter_length
        self.movie.chapter_location = res.field_chapter_location

    async def get_content_details(
        self, handle: str, passcode: str = None
    ) -> messages.ContentDetailsOverview:
        """Return content details for the currently selected title."""
        responses: list[Response] = await self._send_multi(
            messages.GetContentDetails, 0, [handle, passcode if passcode else ""]
        )
        overview = cast(messages.ContentDetailsOverview, responses[0])
        for response in responses[1:]:
            overview.details.update(cast(messages.ContentDetails, response).field)
        return overview

    def _update_content_details(
        self, res: messages.ContentDetailsOverview = None
    ) -> None:
        self.movie.handle = res.field_handle if res else ""
        self.movie.title = res.field_title if res else ""
        self.movie.cover = res.field_cover_url if res else ""
        self.movie.cover_hires = res.field_hires_cover_url if res else ""
        self.movie.rating = res.field_rating if res else ""
        self.movie.rating_reason = res.field_rating_reason if res else ""
        self.movie.year = res.field_year if res else ""
        self.movie.runtime = res.field_running_time if res else "0"
        self.movie.actors = res.field_actors if res else []
        self.movie.director = res.field_director if res else ""
        self.movie.directors = res.field_directors if res else []
        self.movie.genre = res.field_genre if res else ""
        self.movie.genres = res.field_genres if res else []
        self.movie.synopsis = res.field_synopsis if res else ""
        self.movie.color = res.field_color_description if res else ""
        self.movie.country = res.field_country if res else ""
        self.movie.aspect_ratio = res.field_aspect_ratio if res else ""

    async def _get_movie_location(self) -> messages.GetMovieLocation:
        """Return movie location."""
        res = await self._send(messages.GetMovieLocation)
        return cast(messages.GetMovieLocation, res)

    def _update_movie_location(self, res: messages.MovieLocation) -> None:
        self.automation.movie_location = res.field

    async def _get_movie_media_type(self) -> messages.MovieMediaType:
        """Return movie media type."""
        res = await self._send(messages.GetMovieMediaType)
        return cast(messages.MovieMediaType, res)

    def _update_movie_media_type(self, res: messages.MovieMediaType) -> None:
        self.movie.media_type = res.field

    async def _get_video_color(self) -> messages.VideoColor:
        """Return video color."""
        res = await self._send(messages.GetVideoColor)
        return cast(messages.VideoColor, res)

    def _update_video_color(self, res: messages.VideoColor) -> None:
        self.automation.video_color_eotf = res.field_eotf
        self.automation.video_color_space = res.field_space
        self.automation.video_color_depth = res.field_depth
        self.automation.video_color_sampling = res.field_sampling

    async def _get_video_mode(self) -> messages.VideoMode:
        """Return video mode."""
        res = await self._send(messages.GetVideoMode)
        return cast(messages.VideoMode, res)

    def _update_video_mode(self, res: messages.VideoMode) -> None:
        self.automation.video_mode = res.field

    async def _get_screen_mask(self) -> messages.ScreenMask:
        """Return screen mask."""
        res = await self._send(messages.GetScreenMask)
        return cast(messages.ScreenMask, res)

    def _update_screen_mask(self, res: messages.ScreenMask) -> None:
        self.automation.screen_mask_ratio = res.field_image_ratio
        self.automation.screen_mask_top_trim_rel = res.field_top_trim_rel
        self.automation.screen_mask_bottom_trim_rel = res.field_bottom_trim_rel
        self.automation.screen_mask_conservative_ratio = res.field_conservative_ratio
        self.automation.screen_mask_top_mask_abs = res.field_top_mask_abs
        self.automation.screen_mask_bottom_mask_abs = res.field_bottom_mask_abs

    async def _get_screen_mask2(self) -> messages.ScreenMask2:
        """Return screen mask2."""
        res = await self._send(messages.GetScreenMask2)
        return cast(messages.ScreenMask2, res)

    def _update_screen_mask2(self, res: messages.ScreenMask2) -> None:
        self.automation.screen_mask2_top_mask_abs = res.field_top_mask_abs
        self.automation.screen_mask2_bottom_mask_abs = res.field_bottom_mask_abs
        self.automation.screen_mask2_top_calibrated = res.field_top_calibrated
        self.automation.screen_mask2_bottom_calibrated = res.field_bottom_calibrated

    async def _get_cinemascape_mode(self) -> messages.CinemascapeMode:
        """Return cinemascape mode."""
        res = await self._send(messages.GetCinemascapeMode)
        return cast(messages.CinemascapeMode, res)

    def _update_cinemascape_mode(self, res: messages.CinemascapeMode) -> None:
        self.automation.cinemascape_mode = res.field

    async def _get_cinemascape_mask(self) -> messages.CinemascapeMask:
        """Return cinemascape mask."""
        res = await self._send(messages.GetCinemascapeMask)
        return cast(messages.CinemascapeMask, res)

    def _update_cinemascape_mask(self, res: messages.CinemascapeMask) -> None:
        self.automation.cinemascape_mask = res.field

    async def _send(
        self, request: type[RequestType], zone: int = 0, fields: list[str] | None = None
    ) -> Response:
        """Send request to hardware, returning a single response."""
        res = await self._send_multi(request, zone, fields)
        assert len(res) == 1
        return res[0]

    async def _send_multi(
        self, request: type[RequestType], zone: int = 0, fields: list[str] | None = None
    ) -> list[Response]:
        """Send request to hardware, returning one or more responses."""
        req = request(zone, fields)
        return await req.send(self._connection)

    async def _handle_event(self, response: Response) -> None:
        """Handle events sent by hardware."""

        # System
        if isinstance(response, messages.DevicePowerState):
            self._update_device_power_state(response)
            await self.refresh()
        elif isinstance(response, messages.SystemReadinessState):
            self._update_system_readiness_state(response)
        elif isinstance(response, messages.FriendlyName):
            self._update_friendly_name(response)

        # OSD
        elif isinstance(response, messages.UiState):
            self._update_ui_state(response)
        elif isinstance(response, messages.PlayingTitleName):
            self._update_playing_title_name(response)
        elif isinstance(response, messages.HighlightedSelection):
            self._update_highlighted_selection(response)

        # Movie
        elif isinstance(response, messages.PlayStatus):
            old_mode = self.movie.play_status
            self._update_play_status(response)
            if (
                self.power.state == const.DEVICE_POWER_STATE_ON
                and self.movie.play_status != const.PLAY_STATUS_NONE
            ):
                if old_mode == const.PLAY_STATUS_NONE or old_mode is None:
                    res = await self.get_content_details(self.osd.highlighted)
                    self._update_content_details(
                        cast(messages.ContentDetailsOverview, res)
                    )
            elif self.movie.title:
                self._update_content_details()
        elif isinstance(response, messages.MovieMediaType):
            self._update_movie_media_type(response)

        # Automation
        elif isinstance(response, messages.MovieLocation):
            self._update_movie_location(response)
        elif isinstance(response, messages.VideoColor):
            self._update_video_color(response)
        elif isinstance(response, messages.VideoMode):
            self._update_video_mode(response)
        elif isinstance(response, messages.ScreenMask):
            self._update_screen_mask(response)
        elif isinstance(response, messages.ScreenMask2):
            self._update_screen_mask2(response)
        elif isinstance(response, messages.CinemascapeMode):
            self._update_cinemascape_mode(response)
        elif isinstance(response, messages.CinemascapeMask):
            self._update_cinemascape_mask(response)

        self._dispatcher.send(response.name)

    @property
    def dispatcher(self) -> Dispatcher:
        """Return dispatcher instance."""
        return self._dispatcher

    @property
    def connection(self) -> Connection:
        """Return connection instance."""
        return self._connection

    @property
    def host(self) -> str:
        """Return connection host."""
        return self._host

    @property
    def port(self) -> int:
        """Return connection port."""
        return self._port

    @property
    def serial_number(self) -> str:
        """Return hardware's serial number."""
        return self.system.serial_number

    @property
    def is_connected(self) -> bool:
        """Return current state of the connection."""
        return self._connection.state == const.STATE_CONNECTED

    @property
    def is_server_only(self) -> bool:
        """Return if device has no movie zone (Terra, 1U, 3U, etc)."""
        return self.system.movie_zones == 0

    @property
    def is_movie_player(self) -> bool:
        """Return if device has a movie zone."""
        return self.system.movie_zones > 0

    @property
    def is_music_player(self) -> bool:
        """Return if device has a music zone."""
        return self.system.music_zones - self.system.movie_zones > 0


@dataclass
class System:
    """System related properties."""

    ip_address: str = ""
    serial_number: str = ""
    cpdid: str = ""
    type: str = ""
    protocol: int = 0
    kos_version: str = ""
    friendly_name: str = ""
    movie_zones: int = 0
    music_zones: int = 0


@dataclass
class Power:
    """Power related state."""

    state: str = ""
    readiness: str = ""
    zone: list[str] | None = None


@dataclass
class OSD:
    """On Screen Display related state."""

    ui_screen: str = const.UI_STATE_SCREEN_UNKNOWN
    ui_popup: str = const.UI_STATE_POPUP_NONE
    ui_dialog: str = const.UI_STATE_DIALOG_NONE
    ui_screensaver: str = const.UI_STATE_SAVER_INACTIVE
    title_name: str = ""
    highlighted: str = ""


@dataclass
class Movie:
    """Movie media related state."""

    handle: str = ""
    title: str = ""
    cover: str = ""
    cover_hires: str = ""
    rating: str = ""
    rating_reason: str = ""
    year: str = ""
    runtime: str = ""
    actors: list[str] | None = None
    director: str = ""
    directors: list[str] | None = None
    genre: str = ""
    genres: list[str] | None = None
    synopsis: str = ""
    color: str = ""
    country: str = ""
    aspect_ratio: str = ""
    media_type: str = const.MOVIE_MEDIA_TYPE_NONE
    play_status: str = const.PLAY_STATUS_NONE
    play_speed: int = 0
    title_number: int = 0
    title_length: int = 0
    title_location: int = 0
    chapter_number: int = 0
    chapter_length: int = 0
    chapter_location: int = 0


@dataclass
class Automation:
    """Automation related state."""

    movie_location: str = const.MOVIE_LOCATION_NONE
    cinemascape_mask: int = 0
    cinemascape_mode: str = const.CINEMASCAPE_MODE_NONE
    video_mode: str = const.VIDEO_MODE_NONE
    video_color_eotf: str = const.VIDEO_COLOR_EOTF_UNKNOWN
    video_color_space: str = const.VIDEO_COLOR_SPACE_DEFAULT
    video_color_depth: str = const.VIDEO_COLOR_DEPTH_UNKNOWN
    video_color_sampling: str = const.VIDEO_COLOR_SAMPLING_NONE
    screen_mask_ratio: str = const.SCREEN_MASK_ASPECT_RATIO_NONE
    screen_mask_top_trim_rel: int = 0
    screen_mask_bottom_trim_rel: int = 0
    screen_mask_conservative_ratio: str = ""
    screen_mask_top_mask_abs: int = 0
    screen_mask_bottom_mask_abs: int = 0
    screen_mask2_top_mask_abs: int = 0
    screen_mask2_bottom_mask_abs: int = 0
    screen_mask2_top_calibrated: int = 0
    screen_mask2_bottom_calibrated: int = 0



"""Classes for dispatching events"""

from __future__ import annotations

import asyncio
import functools
import logging
from collections.abc import Awaitable, Callable
from typing import Any

_LOGGER = logging.getLogger(__name__)


class Signal:
    """Container for a named target function that receives events"""

    def __init__(self, dispatcher: Dispatcher, target: Callable):
        """Initialize signal."""
        self.dispatcher = dispatcher
        self.target = target

    def disconnect(self) -> None:
        """Removes signal from the dispatcher."""
        self.dispatcher.disconnect(self)


class Dispatcher:
    """Handle event dispatching."""

    def __init__(self):
        """Initialize dispatcher."""
        self._signals: list[Signal] = []
        self._loop = asyncio.get_event_loop()
        self._disconnects = []

    def connect(self, target: Callable) -> Signal:
        """Return a new signal that runs the target function."""
        signal = Signal(self, target)
        self._signals.append(signal)
        return signal

    def send(self, *args: Any) -> None:
        """Call named signal's target function with args."""
        for signal in self._signals:
            self._call_target(signal.target, *args)
        if len(self._signals) > 0:
            _LOGGER.debug(
                "Dispatched signal to %s listener%s with %s",
                len(self._signals),
                "s" if len(self._signals) > 1 else "",
                args,
            )

    def disconnect(self, signal: Signal):
        """Removes signal."""
        try:
            self._signals.remove(signal)
        except ValueError:
            pass

    def disconnect_all(self) -> None:
        """Disconnect all signals."""
        self._signals.clear()

    def _call_target(self, target: Callable, *args) -> Awaitable | None:
        check_target = target
        while isinstance(check_target, functools.partial):
            check_target = check_target.func
        if asyncio.iscoroutinefunction(check_target):
            return self._loop.create_task(target(*args))
        self._loop.call_soon(target, *args)



"""Tests for message module."""

import pytest

from kaleidescape import const
from kaleidescape import message as messages
from kaleidescape.error import MessageParseError
from kaleidescape.message import MESSAGE_TYPE_REQUEST, MessageParser, Request, Response


def test_message_parser():
    """Test message parser."""
    parsed = MessageParser("01/2/003:MESSAGE_NAME:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    assert parsed.seq == 2
    assert parsed.status == 3
    assert parsed.name == "MESSAGE_NAME"
    assert parsed.fields == []
    assert parsed.checksum == 123


def test_message_parser_with_fields():
    """Test message parser with fields."""
    parsed = MessageParser("01/2/003:MESSAGE_NAME:field1:field2:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    assert parsed.seq == 2
    assert parsed.status == 3
    assert parsed.name == "MESSAGE_NAME"
    assert parsed.fields == ["field1", "field2"]
    assert parsed.checksum == 123


def test_message_parser_event():
    """Test message parse as event."""
    parsed = MessageParser("01/!/003:MESSAGE_NAME:field1:field2:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    assert parsed.seq == -1
    assert parsed.status == 3
    assert parsed.name == "MESSAGE_NAME"
    assert parsed.fields == ["field1", "field2"]
    assert parsed.checksum == 123


def test_message_parser_request():
    """Test message parse when message is request."""
    parsed = MessageParser("01/1/GET_MESSAGE_NAME:field1:field2:/123", True)
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    assert parsed.seq == 1
    assert parsed.status == 0
    assert parsed.name == "GET_MESSAGE_NAME"
    assert parsed.fields == ["field1", "field2"]
    assert parsed.checksum is 0


def test_message_parser_device_id():
    """Test message parser device_id."""
    parsed = MessageParser("01/2/003:MESSAGE_NAME:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    parsed = MessageParser("01.02/2/003:MESSAGE_NAME:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 2
    parsed = MessageParser("#0A123F12/2/003:MESSAGE_NAME:/123")
    assert parsed.device_id == "#0A123F12"
    assert parsed.zone == 0
    parsed = MessageParser("#0A123F12.02/2/003:MESSAGE_NAME:/123")
    assert parsed.device_id == "#0A123F12"
    assert parsed.zone == 2


def test_message_parser_bad_device_id():
    """Test message parser with bad device_id."""
    with pytest.raises(MessageParseError) as err:
        MessageParser("1/2/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("001/2/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("aa/2/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("/2/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01./2/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01.0a/2/003:MESSAGE_NAME:/123")


def test_message_parser_seq():
    """Test message parser seq."""
    parsed = MessageParser("01/9/003:MESSAGE_NAME:/123")
    assert parsed.seq == 9
    parsed = MessageParser("01/!/003:MESSAGE_NAME:/123")
    assert parsed.seq == -1


def test_message_parser_bad_seq():
    """Test message parser with bad seq."""
    with pytest.raises(MessageParseError) as err:
        MessageParser("01//003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_SEQ_NUMBER] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/22/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_SEQ_NUMBER] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/aa/003:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_SEQ_NUMBER] in str(err.value)


def test_message_parser_status():
    """Test message parser status."""
    parsed = MessageParser("01/0/999:MESSAGE_NAME:/123")
    assert parsed.status == 999


def test_message_parser_bad_status():
    """Test message parser with bad status."""
    with pytest.raises(MessageParseError) as err:
        MessageParser("01/2/0001:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_UNDETERMINED_ERROR] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/2/01:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_UNDETERMINED_ERROR] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/2/aaa:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_UNDETERMINED_ERROR] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/2/:MESSAGE_NAME:/123")
    assert const.RESPONSE_ERROR[const.ERROR_UNDETERMINED_ERROR] in str(err.value)


def test_message_parser_name():
    """Test message parser name."""
    parsed = MessageParser("01/2/003:MESSAGE_NAME_123:/123")
    assert parsed.name == "MESSAGE_NAME_123"


def test_message_parser_bad_name():
    """Test message parser with bad name."""
    with pytest.raises(MessageParseError) as err:
        MessageParser("01/2/003:MESSAGE_NAME/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_REQUEST] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/2//123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_REQUEST] in str(err.value)
    with pytest.raises(MessageParseError):
        MessageParser("01/2/:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_REQUEST] in str(err.value)


def test_message_parser_field_decoding():
    """Test message parser field decoding."""
    parsed = MessageParser("01/2/003:MESSAGE_NAME:f\\d225ncy:/123")
    assert parsed.fields == ["fáncy"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:f\\\\d225ncy:/123")
    assert parsed.fields == ["f\\d225ncy"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:colon\\:separated:/123")
    assert parsed.fields == ["colon:separated"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:forward\\/slash:/123")
    assert parsed.fields == ["forward/slash"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:back\\\\slash:/123")
    assert parsed.fields == ["back\\slash"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:new\\nline:/123")
    assert parsed.fields == ["new\nline"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:new\\\nline:/123")
    assert parsed.fields == ["new\nline"]
    parsed = MessageParser("01/2/003:MESSAGE_NAME:new\\\rline:/123")
    assert parsed.fields == ["new\rline"]


def test_message_parser_bad_fields():
    """Test message parser with bad fields."""
    with pytest.raises(MessageParseError):
        MessageParser("01/2/:MESSAGE_NAME:field/123")


def test_message_parser_empty_name():
    """Test message parser with empty name."""
    parsed = MessageParser("01/2/000:/123")
    assert parsed.device_id == "01"
    assert parsed.zone == 0
    assert parsed.seq == 2
    assert parsed.status == 0
    assert parsed.name == ""
    assert parsed.fields == []
    assert parsed.checksum == 123


def test_message_response_with_empty_name():
    """Test message response with empty name"""
    message = Response(MessageParser("01/2/999:/123"))
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == 2
    assert message.status == 999
    assert message.fields == []
    assert message.is_event is False
    assert message.is_error is True
    assert message.error == const.RESPONSE_ERROR[message.status]
    assert str(message) == "01/2/999:/123"


def test_message_response_from_command():
    """Test message response from command."""
    message = Response.factory("01/2/000:AVAILABLE_DEVICES:01:02:/123")
    assert isinstance(message, messages.AvailableDevices)
    assert message.type == messages.MESSAGE_TYPE_RESPONSE
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == 2
    assert message.status == 0
    assert message.name == "AVAILABLE_DEVICES"
    assert message.fields == ["01", "02"]
    assert message.is_event is False
    assert message.is_error is False
    assert str(message) == "01/2/000:AVAILABLE_DEVICES:01:02:/123"


def test_message_response_from_event():
    """Test message response from event."""
    message = Response.factory("01/!/000:AVAILABLE_DEVICES:01:02:/123")
    assert isinstance(message, messages.AvailableDevices)
    assert message.type == messages.MESSAGE_TYPE_EVENT
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == -1
    assert message.status == 0
    assert message.name == "AVAILABLE_DEVICES"
    assert message.fields == ["01", "02"]
    assert message.is_event is True
    assert message.is_error is False
    assert str(message) == "01/!/000:AVAILABLE_DEVICES:01:02:/123"


def test_message_response_with_error_status():
    """Test message response with error status."""
    message = Response.factory("01/2/003:AVAILABLE_DEVICES:01:02:/123")
    assert isinstance(message, messages.AvailableDevices)
    assert message.type == messages.MESSAGE_TYPE_RESPONSE
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == 2
    assert message.status == 3
    assert message.name == "AVAILABLE_DEVICES"
    assert message.fields == ["01", "02"]
    assert message.is_event is False
    assert message.is_error is True
    assert message.error == const.RESPONSE_ERROR[message.status]
    assert str(message) == "01/2/003:AVAILABLE_DEVICES:01:02:/123"


def test_message_response_with_parse_error():
    """Test message response with parse error."""
    with pytest.raises(MessageParseError) as err:
        Response.factory("ERR/2/003:AVAILABLE_DEVICES:01:02:/123")
    assert const.RESPONSE_ERROR[const.ERROR_INVALID_DEVICE] in str(err.value)


def test_message_request():
    """Test message request."""
    message = Request()
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == -1
    assert message.status == 0
    assert message.fields == []
    assert message.type == MESSAGE_TYPE_REQUEST


def test_message_request_with_fields():
    """Test message request."""
    message = Request(0, ["field1", "field2"])
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == -1
    assert message.status == 0
    assert message.fields == ["field1", "field2"]
    assert message.type == MESSAGE_TYPE_REQUEST


def test_message_get_available_devices():
    """Test message request as GET_AVAILABLE_DEVICES."""
    message = messages.GetAvailableDevices()
    message.seq = 2
    assert message.device_id == "01"
    assert message.zone == 0
    assert message.seq == 2
    assert message.status == 0
    assert message.name == "GET_AVAILABLE_DEVICES"
    assert message.fields == []
    assert str(message) == "01/2/GET_AVAILABLE_DEVICES:"



"""Simple Kaleidescape Control Protocol emulator used for unit testing."""

from __future__ import annotations

import asyncio
import logging

from kaleidescape import const, error
from kaleidescape import message as messages
from kaleidescape.connection import SEPARATOR
from kaleidescape.const import SUCCESS
from kaleidescape.message import MessageParser

_LOGGER = logging.getLogger(__name__)


# pylint: disable=missing-function-docstring, invalid-name, no-self-use, line-too-long
# fmt: off

# Returns the inverse of map
def xo(val: dict[int, str]) -> dict[str, int]:
    return {v: k for k, v in val.items()}


class Request:
    """Class representing a request sent to Emulator from Client."""

    def __init__(self, msg: str) -> None:
        parsed = MessageParser(msg, True)
        self.message = msg
        self.device_id: str = parsed.device_id
        self.seq: int = parsed.seq
        self.name: str = parsed.name
        self.fields: list[str] = parsed.fields
        self.checksum: int = parsed.checksum


class Response:
    """Class representing a response sent from Emulator to Client."""

    def __init__(
        self,
        device_id: str | None,
        seq: int = 0,
        status: int = 0,
        name: str = None,
        fields: list[str] = None,
    ):
        self.device_id = device_id
        self.seq = seq
        self.status = status
        self.name = name
        self.fields = fields
        self.checksum: int = 1

    def __str__(self) -> str:
        msg = [f"{self.status:03}"]
        if self.name:
            msg.append(self.name)
        if self.fields:
            fields = [self._encode(str(f)) for f in self.fields]
            msg = msg + fields
        seq = "!" if self.seq < 0 else str(self.seq)
        return f"{self.device_id}/{seq}/{':'.join(msg)}:/1"

    def _encode(self, field: str) -> str:
        field = field.replace("\\", "\\\\")
        field = field.replace("/", r"\/")
        field = field.replace("\n", "\\\n")  # The escaped newline emulated the bug found in actual devices
        field = field.replace("\r", "\\\r")
        field = field.replace("\t", r"\t")
        field = field.replace(":", r"\:")
        for i in range(192, 255):
            field = field.replace(chr(i), "\\d{i:03}")
        return field


class Event(Response):
    """Class for responses that are broadcast to multiple Devices."""

    def __init__(self, source_device_ids: list[str], status: int, name: str, fields: list = None) -> None:
        super().__init__(None, -1, status, name, fields)
        self.source_device_ids = source_device_ids


class Client:
    """Class representing a connection to Emulator."""

    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self._reader = reader
        self._writer = writer

    async def disconnect(self):
        self._reader.feed_eof()
        self._writer.close()

    async def _write(self, payload: str):
        data = (payload + SEPARATOR).encode("latin-1")
        self._writer.write(data)
        await self._writer.drain()

    async def send(self, msg: str | Response | Event):
        if isinstance(msg, Event):
            for device_id in msg.source_device_ids:
                msg.device_id = device_id
                await self._write(str(msg))
                _LOGGER.debug("> %s", str(msg))
        elif isinstance(msg, Response):
            await self._write(str(msg))
            _LOGGER.debug("> %s", str(msg))
        else:
            await self._write(msg)
            _LOGGER.debug("> %s", msg)


class Emulator:
    """Class for emulating a Kaleidescape system."""

    def __init__(self, host: str, port: int = const.DEFAULT_PROTOCOL_PORT):
        """Initialize the emulator."""
        self._host = host
        self._port = port
        self._clients: list[Client] = []
        self._server: asyncio.base_events.Server | None = None
        self._mock_commands: dict[str, dict] = {}

        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetAvailableDevicesBySerialNumber.name,
            (SUCCESS, messages.AvailableDevicesBySerialNumber.name, ["00000000123A"])
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetAvailableDevices.name,
            (SUCCESS, messages.AvailableDevices.name, ["01"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetDeviceInfo.name,
            (SUCCESS, messages.DeviceInfo.name, ["", "00000000123A", "00", "127.0.0.1"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetSystemVersion.name,
            (SUCCESS, messages.SystemVersion.name, ["16", "10.4.2-19218"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetNumZones.name,
            (SUCCESS, messages.NumZones.name, ["01", "01"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetDeviceTypeName.name,
            (SUCCESS, messages.DeviceTypeName.name, ["Strato S"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetFriendlyName.name,
            (SUCCESS, messages.FriendlyName.name, ["Theater"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetDevicePowerState.name,
            (SUCCESS, messages.DevicePowerState.name, ["0", "0"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetSystemReadinessState.name,
            (SUCCESS, messages.SystemReadinessState.name, ["2"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetContentDetails.name,
            [
                (SUCCESS, messages.ContentDetailsOverview.name, ["2", "26-0.0-S_c446c8e2", "movies"]),
                (SUCCESS, messages.ContentDetails.name, ["1", "Content_handle", "26-0.0-S_c446c8e2"]),
                (SUCCESS, messages.ContentDetails.name, ["2", "Title", "Turtle Odyssey"]),
            ]
        )

        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetFriendlySystemName.name,
            (SUCCESS, messages.FriendlySystemName.name, ["Home Cinema"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetZoneCapabilities.name,
            (SUCCESS, messages.ZoneCapabilities.name, ["Y", "Y", "N", "Y"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetUiState.name,
            (SUCCESS, messages.UiState.name, ["01", "00", "00", "0"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetHighlightedSelection.name,
            (SUCCESS, messages.HighlightedSelection.name, [""]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetPlayStatus.name,
            (SUCCESS, messages.PlayStatus.name, ["0", "1", "00", "00000", "00000", "000", "00000", "00000"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetMovieLocation.name,
            (SUCCESS, messages.MovieLocation.name, ["03"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetVideoColor.name,
            (SUCCESS, messages.VideoColor.name, ["00", "00", "24", "00"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetVideoMode.name,
            (SUCCESS, messages.VideoMode.name, ["00", "00", "00"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetScreenMask.name,
            (SUCCESS, messages.ScreenMask.name, ["00", "000", "000", "05", "0000", "0000"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetScreenMask2.name,
            (SUCCESS, messages.ScreenMask2.name, ["00", "00", "00000", "00000"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetCinemascapeMode.name,
            (SUCCESS, messages.CinemascapeMode.name, ["0"]),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.GetCinemascapeMask.name,
            (SUCCESS, messages.CinemascapeMask.name, ["000"]),
        )

        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.EnableEvents.name,
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.LeaveStandby.name,
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            messages.EnterStandby.name,
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            "PLAY",  # Fake command for simulating slow work
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            "PAUSE",  # Fake command for simulating slow work
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            "STOP",  # Fake command for simulating slow work
            (SUCCESS,),
        )
        self.register_mock_command(
            ("01", "#00000000123A"),
            "_SLEEP",  # Fake command for simulating slow work
            (SUCCESS,),
        )

    async def start(self):
        """Starts the emulator."""
        if self._server:
            raise Exception("Already started")
        self._server = await asyncio.start_server(
            self._connection_handler, self._host, self._port
        )
        _LOGGER.debug("Started")

    async def stop(self):
        """Stops the emulator."""
        if self._server is None:
            return
        for client in self._clients:
            await client.disconnect()
        self._clients.clear()
        self._server.close()
        await self._server.wait_closed()
        self._server = None
        await asyncio.sleep(0.01)
        _LOGGER.debug("Stopped")

    def register_mock_command(
            self,
            device_ids: tuple[str, ...],
            name: str,
            msg: tuple | list
    ):
        """Adds a new simulated command to server. Overrides built in commands."""
        for device_id in device_ids:
            self._mock_commands.setdefault(device_id, {})
            self._mock_commands[device_id][name] = msg

    def change_mock_device_id(self, old_cpdid, new_cpdid):
        self._mock_commands[new_cpdid] = dict(self._mock_commands[old_cpdid])
        del self._mock_commands[old_cpdid]

    def unregister_mock_device_ids(self, device_ids: tuple[str, ...]):
        """Removes all simulated commands for device_id."""
        for device_id in device_ids:
            del self._mock_commands[device_id]

    async def _connection_handler(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """Main service loop for handling client connections."""
        client = Client(reader, writer)
        self._clients.append(client)
        _LOGGER.debug("Client connected")

        while True:
            request = None
            try:
                result = await reader.readuntil()
                result = result.decode("latin-1").strip()
                if not result:
                    continue

                _LOGGER.debug("< %s", result)

                request = Request(result)

                if request.device_id in self._mock_commands:
                    if request.name in self._mock_commands[request.device_id]:
                        msgs = self._mock_commands[request.device_id][request.name]
                    else:
                        raise error.MessageError(const.ERROR_INVALID_REQUEST, request.message)
                else:
                    raise error.MessageError(const.ERROR_INVALID_DEVICE, request.message)

                if isinstance(msgs, int):
                    msgs = [(msgs,)]
                elif isinstance(msgs, tuple):
                    msgs = [msgs]

                if request.name == "_SLEEP":
                    # 01/1/_SLEEP:1:  # Sleep 1s
                    await asyncio.sleep(float(request.fields[0]))

                for msg in msgs:
                    await client.send(Response(request.device_id, request.seq, *msg))

                await asyncio.sleep(0.001)

            except asyncio.IncompleteReadError:
                # Occurs when the reader is being stopped
                break
            except (error.MessageError, error.MessageParseError) as e:
                device_id = request.device_id if request else "??"
                seq = request.seq if request else "?"
                response = Response(device_id, seq, e.code, e.error)
                await client.send(response)

        try:
            self._clients.remove(client)
            _LOGGER.debug("Client disconnected")
        except ValueError:
            pass

    async def send_event(self, device_ids: list[str], status: int, name: str, fields: list = None):
        """Sends an event message to device_ids."""
        event = Event(device_ids, status, name, fields)
        for client in self._clients:
            await client.send(event)



"""A python client library for controlling Kaleidescape devices via the Kaleidescape
Control Protocol."""

from . import const
from .device import Device
from .dispatcher import Dispatcher
from .error import KaleidescapeError

__version__ = "1.0.0"
